"""Phono3py command option and conf file parser."""
# Copyright (C) 2015 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from phonopy.cui.settings import ConfParser, Settings, fracval


class Phono3pySettings(Settings):
    """Setting parameter container."""

    _default = {
        # In micrometre. The default value is just set to avoid divergence.
        "boundary_mfp": 1.0e6,
        "conductivity_type": None,
        "constant_averaged_pp_interaction": None,
        "create_forces_fc2": None,
        "create_forces_fc3": None,
        "create_forces_fc3_file": None,
        "cutoff_fc3_distance": None,
        "cutoff_pair_distance": None,
        "grid_addresses": None,
        "grid_points": None,
        "grid_matrix": None,
        "ion_clamped": False,
        "is_bterta": False,
        "is_compact_fc": False,
        "is_fc3_r0_average": False,
        "is_full_pp": False,
        "is_gruneisen": False,
        "is_imag_self_energy": False,
        "is_isotope": False,
        "is_joint_dos": False,
        "is_kappa_star": True,
        "is_lbte": False,
        "is_N_U": False,
        "is_real_self_energy": False,
        "is_reducible_collision_matrix": False,
        "is_spectral_function": False,
        "is_symmetrize_fc2": False,
        "is_symmetrize_fc3_q": False,
        "is_symmetrize_fc3_r": False,
        "is_tetrahedron_method": False,
        "lapack_zheev_uplo": "L",
        "mass_variances": None,
        "max_freepath": None,
        "num_points_in_batch": None,
        "read_collision": None,
        "read_fc2": False,
        "read_fc3": False,
        "read_gamma": False,
        "read_phonon": False,
        "read_pp": False,
        "output_yaml_filename": None,
        "phonon_supercell_matrix": None,
        "pinv_cutoff": 1.0e-8,
        "pinv_solver": 0,
        "pinv_method": 0,
        "pp_conversion_factor": None,
        "scattering_event_class": None,  # scattering event class 1 or 2
        "sigma_cutoff_width": None,
        "solve_collective_phonon": False,
        "emulate_v2": False,
        "subtract_forces": None,
        "temperatures": None,
        "use_ave_pp": False,
        "use_grg": False,
        "write_collision": False,
        "write_gamma_detail": False,
        "write_gamma": False,
        "write_phonon": False,
        "write_pp": False,
        "write_LBTE_solution": False,
    }

    def __init__(self, default=None):
        """Init method.

        self._v comes from the parent class.

        """
        super().__init__(default=default)
        self._v.update(Phono3pySettings._default.copy())
        if default is not None:
            self._v.update(default)

    def set_boundary_mfp(self, val):
        """Set boundary_mfp."""
        self._v["boundary_mfp"] = val

    def set_conductivity_type(self, val):
        """Set conductivity_type."""
        self._v["conductivity_type"] = val

    def set_constant_averaged_pp_interaction(self, val):
        """Set constant_averaged_pp_interaction."""
        self._v["constant_averaged_pp_interaction"] = val

    def set_create_forces_fc2(self, val):
        """Set create_forces_fc2."""
        self._v["create_forces_fc2"] = val

    def set_create_forces_fc3(self, val):
        """Set create_forces_fc3."""
        self._v["create_forces_fc3"] = val

    def set_create_forces_fc3_file(self, val):
        """Set create_forces_fc3_file."""
        self._v["create_forces_fc3_file"] = val

    def set_cutoff_fc3_distance(self, val):
        """Set cutoff_fc3_distance."""
        self._v["cutoff_fc3_distance"] = val

    def set_cutoff_pair_distance(self, val):
        """Set cutoff_pair_distance."""
        self._v["cutoff_pair_distance"] = val

    def set_emulate_v2(self, val):
        """Set emulate_v2."""
        self._v["emulate_v2"] = val

    def set_grid_addresses(self, val):
        """Set grid_addresses."""
        self._v["grid_addresses"] = val

    def set_grid_points(self, val):
        """Set grid_points."""
        self._v["grid_points"] = val

    # def set_grid_matrix(self, val):
    #     """Set grid_matrix."""
    #     self._v["grid_matrix"] = val

    def set_ion_clamped(self, val):
        """Set ion_clamped."""
        self._v["ion_clamped"] = val

    def set_is_bterta(self, val):
        """Set is_bterta."""
        self._v["is_bterta"] = val

    def set_is_compact_fc(self, val):
        """Set is_compact_fc."""
        self._v["is_compact_fc"] = val

    def set_is_fc3_r0_average(self, val):
        """Set is_fc3_r0_average."""
        self._v["is_fc3_r0_average"] = val

    def set_is_full_pp(self, val):
        """Set is_full_pp."""
        self._v["is_full_pp"] = val

    def set_is_gruneisen(self, val):
        """Set is_gruneisen."""
        self._v["is_gruneisen"] = val

    def set_is_imag_self_energy(self, val):
        """Set is_imag_self_energy."""
        self._v["is_imag_self_energy"] = val

    def set_is_isotope(self, val):
        """Set is_isotope."""
        self._v["is_isotope"] = val

    def set_is_joint_dos(self, val):
        """Set is_joint_dos."""
        self._v["is_joint_dos"] = val

    def set_is_kappa_star(self, val):
        """Set is_kappa_star."""
        self._v["is_kappa_star"] = val

    def set_is_lbte(self, val):
        """Set is_lbte."""
        self._v["is_lbte"] = val

    def set_is_N_U(self, val):
        """Set is_N_U."""
        self._v["is_N_U"] = val

    def set_is_real_self_energy(self, val):
        """Set is_real_self_energy."""
        self._v["is_real_self_energy"] = val

    def set_is_reducible_collision_matrix(self, val):
        """Set is_reducible_collision_matrix."""
        self._v["is_reducible_collision_matrix"] = val

    def set_is_spectral_function(self, val):
        """Set is_spectral_function."""
        self._v["is_spectral_function"] = val

    def set_is_symmetrize_fc2(self, val):
        """Set is_symmetrize_fc2."""
        self._v["is_symmetrize_fc2"] = val

    def set_is_symmetrize_fc3_q(self, val):
        """Set is_symmetrize_fc3_q."""
        self._v["is_symmetrize_fc3_q"] = val

    def set_is_symmetrize_fc3_r(self, val):
        """Set is_symmetrize_fc3_r."""
        self._v["is_symmetrize_fc3_r"] = val

    def set_lapack_zheev_uplo(self, val):
        """Set lapack_zheev_uplo."""
        self._v["lapack_zheev_uplo"] = val

    def set_mass_variances(self, val):
        """Set mass_variances."""
        self._v["mass_variances"] = val

    def set_max_freepath(self, val):
        """Set max_freepath."""
        self._v["max_freepath"] = val

    def set_num_points_in_batch(self, val):
        """Set num_points_in_batch."""
        self._v["num_points_in_batch"] = val

    def set_output_yaml_filename(self, val):
        """Set output_yaml_filename."""
        self._v["output_yaml_filename"] = val

    def set_phonon_supercell_matrix(self, val):
        """Set phonon_supercell_matrix."""
        self._v["phonon_supercell_matrix"] = val

    def set_pinv_cutoff(self, val):
        """Set pinv_cutoff."""
        self._v["pinv_cutoff"] = val

    def set_pinv_method(self, val):
        """Set pinv_method."""
        self._v["pinv_method"] = val

    def set_pinv_solver(self, val):
        """Set pinv_solver."""
        self._v["pinv_solver"] = val

    def set_pp_conversion_factor(self, val):
        """Set pp_conversion_factor."""
        self._v["pp_conversion_factor"] = val

    def set_read_collision(self, val):
        """Set read_collision."""
        self._v["read_collision"] = val

    def set_read_fc2(self, val):
        """Set read_fc2."""
        self._v["read_fc2"] = val

    def set_read_fc3(self, val):
        """Set read_fc3."""
        self._v["read_fc3"] = val

    def set_read_gamma(self, val):
        """Set read_gamma."""
        self._v["read_gamma"] = val

    def set_read_phonon(self, val):
        """Set read_phonon."""
        self._v["read_phonon"] = val

    def set_read_pp(self, val):
        """Set read_pp."""
        self._v["read_pp"] = val

    def set_scattering_event_class(self, val):
        """Set scattering_event_class."""
        self._v["scattering_event_class"] = val

    def set_sigma_cutoff_width(self, val):
        """Set sigma_cutoff_width."""
        self._v["sigma_cutoff_width"] = val

    def set_solve_collective_phonon(self, val):
        """Set solve_collective_phonon."""
        self._v["solve_collective_phonon"] = val

    def set_subtract_forces(self, val):
        """Set subtract_forces."""
        self._v["subtract_forces"] = val

    def set_temperatures(self, val):
        """Set temperatures."""
        self._v["temperatures"] = val

    def set_use_ave_pp(self, val):
        """Set use_ave_pp."""
        self._v["use_ave_pp"] = val

    def set_use_grg(self, val):
        """Set use_grg."""
        self._v["use_grg"] = val

    def set_write_collision(self, val):
        """Set write_collision."""
        self._v["write_collision"] = val

    def set_write_gamma_detail(self, val):
        """Set write_gamma_detail."""
        self._v["write_gamma_detail"] = val

    def set_write_gamma(self, val):
        """Set write_gamma."""
        self._v["write_gamma"] = val

    def set_write_phonon(self, val):
        """Set write_phonon."""
        self._v["write_phonon"] = val

    def set_write_pp(self, val):
        """Set write_pp."""
        self._v["write_pp"] = val

    def set_write_LBTE_solution(self, val):
        """Set write_LBTE_solution."""
        self._v["write_LBTE_solution"] = val


class Phono3pyConfParser(ConfParser):
    """Parse command options and conf file."""

    def __init__(self, filename=None, args=None, default_settings=None):
        """Init method."""
        self._settings = Phono3pySettings(default=default_settings)
        confs = {}
        if filename is not None:
            super().__init__(filename=filename)
            self.read_file()  # store .conf file setting in self._confs
            self._parse_conf()
            self._set_settings()
            confs.update(self._confs)
        if args is not None:
            super().__init__(args=args)
            self._read_options()
            self._parse_conf()
            self._set_settings()
            confs.update(self._confs)
        self._confs = confs

    def _read_options(self):
        ConfParser.read_options(self)  # store data in self._confs
        if "phonon_supercell_dimension" in self._args:
            dim_fc2 = self._args.phonon_supercell_dimension
            if dim_fc2 is not None:
                self._confs["dim_fc2"] = " ".join(dim_fc2)

        if "boundary_mfp" in self._args:
            if self._args.boundary_mfp is not None:
                self._confs["boundary_mfp"] = self._args.boundary_mfp

        if "const_ave_pp" in self._args:
            const_ave_pp = self._args.const_ave_pp
            if const_ave_pp is not None:
                self._confs["const_ave_pp"] = const_ave_pp

        if "create_forces_fc2" in self._args:
            if self._args.create_forces_fc2:
                self._confs["create_forces_fc2"] = self._args.create_forces_fc2

        if "create_forces_fc3" in self._args:
            if self._args.create_forces_fc3:
                self._confs["create_forces_fc3"] = self._args.create_forces_fc3

        if "create_forces_fc3_file" in self._args:
            if self._args.create_forces_fc3_file:
                cfc3_file = self._args.create_forces_fc3_file
                self._confs["create_forces_fc3_file"] = cfc3_file

        if "cutoff_fc3_distance" in self._args:
            cutoff_fc3 = self._args.cutoff_fc3_distance
            if cutoff_fc3 is not None:
                self._confs["cutoff_fc3_distance"] = cutoff_fc3

        if "cutoff_pair_distance" in self._args:
            cutoff_pair = self._args.cutoff_pair_distance
            if cutoff_pair is not None:
                self._confs["cutoff_pair_distance"] = cutoff_pair

        if "grid_addresses" in self._args:
            grid_adrs = self._args.grid_addresses
            if grid_adrs is not None:
                self._confs["grid_addresses"] = " ".join(grid_adrs)

        if "grid_points" in self._args:
            if self._args.grid_points is not None:
                self._confs["grid_points"] = " ".join(self._args.grid_points)

        if "grid_matrix" in self._args:
            if self._args.grid_matrix is not None:
                self._confs["grid_matrix"] = " ".join(self._args.grid_matrix)

        if "ion_clamped" in self._args:
            if self._args.ion_clamped:
                self._confs["ion_clamped"] = ".true."

        if "is_bterta" in self._args:
            if self._args.is_bterta:
                self._confs["bterta"] = ".true."

        if "is_compact_fc" in self._args:
            if self._args.is_compact_fc:
                self._confs["compact_fc"] = ".true."

        if "emulate_v2" in self._args:
            if self._args.emulate_v2:
                self._confs["emulate_v2"] = ".true."

        if "is_gruneisen" in self._args:
            if self._args.is_gruneisen:
                self._confs["gruneisen"] = ".true."

        if "is_fc3_r0_average" in self._args:
            if self._args.is_fc3_r0_average:
                self._confs["fc3_r0_average"] = ".true."

        if "is_full_pp" in self._args:
            if self._args.is_full_pp:
                self._confs["full_pp"] = ".true."

        if "is_imag_self_energy" in self._args:
            if self._args.is_imag_self_energy:
                self._confs["imag_self_energy"] = ".true."

        if "is_isotope" in self._args:
            if self._args.is_isotope:
                self._confs["isotope"] = ".true."

        if "is_joint_dos" in self._args:
            if self._args.is_joint_dos:
                self._confs["joint_dos"] = ".true."

        if "no_kappa_stars" in self._args:
            if self._args.no_kappa_stars:
                self._confs["kappa_star"] = ".false."

        if "is_lbte" in self._args:
            if self._args.is_lbte:
                self._confs["lbte"] = ".true."

        if "is_N_U" in self._args:
            if self._args.is_N_U:
                self._confs["N_U"] = ".true."

        if "is_real_self_energy" in self._args:
            if self._args.is_real_self_energy:
                self._confs["real_self_energy"] = ".true."

        if "is_reducible_collision_matrix" in self._args:
            if self._args.is_reducible_collision_matrix:
                self._confs["reducible_collision_matrix"] = ".true."

        if "is_spectral_function" in self._args:
            if self._args.is_spectral_function:
                self._confs["spectral_function"] = ".true."

        if "is_symmetrize_fc2" in self._args:
            if self._args.is_symmetrize_fc2:
                self._confs["symmetrize_fc2"] = ".true."

        if "is_symmetrize_fc3_q" in self._args:
            if self._args.is_symmetrize_fc3_q:
                self._confs["symmetrize_fc3_q"] = ".true."

        if "is_symmetrize_fc3_r" in self._args:
            if self._args.is_symmetrize_fc3_r:
                self._confs["symmetrize_fc3_r"] = ".true."

        if "is_tetrahedron_method" in self._args:
            if self._args.is_tetrahedron_method:
                self._confs["tetrahedron"] = ".true."

        if "is_wigner_kappa" in self._args:
            if self._args.is_wigner_kappa:
                self._confs["conductivity_type"] = "wigner"

        if "is_kubo_kappa" in self._args:
            if self._args.is_kubo_kappa:
                self._confs["conductivity_type"] = "kubo"

        if "lapack_zheev_uplo" in self._args:
            if self._args.lapack_zheev_uplo is not None:
                self._confs["lapack_zheev_uplo"] = self._args.lapack_zheev_uplo

        if "mass_variances" in self._args:
            mass_variances = self._args.mass_variances
            if mass_variances is not None:
                self._confs["mass_variances"] = " ".join(mass_variances)

        if "max_freepath" in self._args:
            if self._args.max_freepath is not None:
                self._confs["max_freepath"] = self._args.max_freepath

        if "num_points_in_batch" in self._args:
            num_points_in_batch = self._args.num_points_in_batch
            if num_points_in_batch is not None:
                self._confs["num_points_in_batch"] = num_points_in_batch

        if "output_yaml_filename" in self._args:
            if self._args.output_yaml_filename is not None:
                self._confs["output_yaml_filename"] = self._args.output_yaml_filename

        if "pinv_cutoff" in self._args:
            if self._args.pinv_cutoff is not None:
                self._confs["pinv_cutoff"] = self._args.pinv_cutoff

        if "pinv_method" in self._args:
            if self._args.pinv_method is not None:
                self._confs["pinv_method"] = self._args.pinv_method

        if "pinv_solver" in self._args:
            if self._args.pinv_solver is not None:
                self._confs["pinv_solver"] = self._args.pinv_solver

        if "pp_conversion_factor" in self._args:
            pp_conv_factor = self._args.pp_conversion_factor
            if pp_conv_factor is not None:
                self._confs["pp_conversion_factor"] = pp_conv_factor

        if "read_fc2" in self._args:
            if self._args.read_fc2:
                self._confs["read_fc2"] = ".true."

        if "read_fc3" in self._args:
            if self._args.read_fc3:
                self._confs["read_fc3"] = ".true."

        if "read_gamma" in self._args:
            if self._args.read_gamma:
                self._confs["read_gamma"] = ".true."

        if "read_phonon" in self._args:
            if self._args.read_phonon:
                self._confs["read_phonon"] = ".true."

        if "read_pp" in self._args:
            if self._args.read_pp:
                self._confs["read_pp"] = ".true."

        if "read_collision" in self._args:
            if self._args.read_collision is not None:
                self._confs["read_collision"] = self._args.read_collision

        if "scattering_event_class" in self._args:
            scatt_class = self._args.scattering_event_class
            if scatt_class is not None:
                self._confs["scattering_event_class"] = scatt_class

        if "sigma_cutoff_width" in self._args:
            sigma_cutoff = self._args.sigma_cutoff_width
            if sigma_cutoff is not None:
                self._confs["sigma_cutoff_width"] = sigma_cutoff

        if "solve_collective_phonon" in self._args:
            if self._args.solve_collective_phonon:
                self._confs["collective_phonon"] = ".true."

        if "subtract_forces" in self._args:
            if self._args.subtract_forces:
                self._confs["subtract_forces"] = self._args.subtract_forces

        if "temperatures" in self._args:
            if self._args.temperatures is not None:
                self._confs["temperatures"] = " ".join(self._args.temperatures)

        if "use_ave_pp" in self._args:
            if self._args.use_ave_pp:
                self._confs["use_ave_pp"] = ".true."

        if "use_grg" in self._args:
            if self._args.use_grg:
                self._confs["use_grg"] = ".true."

        if "write_gamma_detail" in self._args:
            if self._args.write_gamma_detail:
                self._confs["write_gamma_detail"] = ".true."

        if "write_gamma" in self._args:
            if self._args.write_gamma:
                self._confs["write_gamma"] = ".true."

        if "write_collision" in self._args:
            if self._args.write_collision:
                self._confs["write_collision"] = ".true."

        if "write_phonon" in self._args:
            if self._args.write_phonon:
                self._confs["write_phonon"] = ".true."

        if "write_pp" in self._args:
            if self._args.write_pp:
                self._confs["write_pp"] = ".true."

        if "write_LBTE_solution" in self._args:
            if self._args.write_LBTE_solution:
                self._confs["write_LBTE_solution"] = ".true."

    def _parse_conf(self):
        ConfParser.parse_conf(self)
        confs = self._confs

        for conf_key in confs.keys():
            # Boolean
            if conf_key in (
                "read_fc2",
                "read_fc3",
                "read_gamma",
                "read_phonon",
                "read_pp",
                "use_ave_pp",
                "use_grg",
                "collective_phonon",
                "write_gamma_detail",
                "write_gamma",
                "write_collision",
                "write_phonon",
                "write_pp",
                "write_LBTE_solution",
                "full_pp",
                "ion_clamped",
                "bterta",
                "compact_fc",
                "emulate_v2",
                "fc3_r0_average",
                "real_self_energy",
                "gruneisen",
                "imag_self_energy",
                "isotope",
                "joint_dos",
                "lbte",
                "N_U",
                "spectral_function",
                "reducible_collision_matrix",
                "symmetrize_fc2",
                "symmetrize_fc3_q",
                "symmetrize_fc3_r",
                "kappa_star",
            ):
                if confs[conf_key].lower() == ".true.":
                    self.set_parameter(conf_key, True)
                elif confs[conf_key].lower() == ".false.":
                    self.set_parameter(conf_key, False)

            # float
            if conf_key in (
                "boundary_mfp",
                "const_ave_pp",
                "cutoff_fc3_distance",
                "cutoff_pair_distance",
                "max_freepath",
                "pinv_cutoff",
                "pp_conversion_factor",
                "sigma_cutoff_width",
            ):
                self.set_parameter(conf_key, float(confs[conf_key]))

            # int
            if conf_key in (
                "pinv_method",
                "pinv_solver",
                "num_points_in_batch",
                "scattering_event_class",
            ):
                self.set_parameter(conf_key, int(confs[conf_key]))

            # string
            if conf_key in (
                "conductivity_type",
                "create_forces_fc3_file",
                "output_yaml_filename",
            ):
                self.set_parameter(conf_key, confs[conf_key])

            # specials
            if conf_key in ("create_forces_fc2", "create_forces_fc3"):
                if type(confs[conf_key]) is str:
                    fnames = confs[conf_key].split()
                else:
                    fnames = confs[conf_key]
                self.set_parameter(conf_key, fnames)

            if conf_key == "dim_fc2":
                matrix = [int(x) for x in confs["dim_fc2"].split()]
                if len(matrix) == 9:
                    matrix = np.array(matrix).reshape(3, 3)
                elif len(matrix) == 3:
                    matrix = np.diag(matrix)
                else:
                    self.setting_error(
                        "Number of elements of dim-fc2 has to be 3 or 9."
                    )

                if matrix.shape == (3, 3):
                    if np.linalg.det(matrix) < 1:
                        self.setting_error(
                            "Determinant of supercell matrix has " + "to be positive."
                        )
                    else:
                        self.set_parameter("dim_fc2", matrix)

            if conf_key == "grid_addresses":
                vals = [
                    int(x) for x in confs["grid_addresses"].replace(",", " ").split()
                ]
                if len(vals) % 3 == 0 and len(vals) > 0:
                    self.set_parameter("grid_addresses", np.reshape(vals, (-1, 3)))
                else:
                    self.setting_error("Grid addresses are incorrectly set.")

            if conf_key == "grid_points":
                vals = [int(x) for x in confs["grid_points"].replace(",", " ").split()]
                self.set_parameter("grid_points", vals)

            if conf_key == "grid_matrix":
                vals = [int(x) for x in confs["grid_matrix"].replace(",", " ").split()]
                if len(vals) == 9:
                    self.set_parameter("grid_matrix", np.reshape(vals, (3, 3)))
                else:
                    self.setting_error("Grid matrix are incorrectly set.")

            if conf_key == "lapack_zheev_uplo":
                self.set_parameter(
                    "lapack_zheev_uplo", confs["lapack_zheev_uplo"].upper()
                )

            if conf_key == "mass_variances":
                vals = [fracval(x) for x in confs["mass_variances"].split()]
                if len(vals) < 1:
                    self.setting_error("Mass variance parameters are incorrectly set.")
                else:
                    self.set_parameter("mass_variances", vals)

            if conf_key == "read_collision":
                if confs["read_collision"] == "all":
                    self.set_parameter("read_collision", "all")
                else:
                    vals = [int(x) for x in confs["read_collision"].split()]
                    self.set_parameter("read_collision", vals)

            # For multiple T values.
            if conf_key == "temperatures":
                vals = [fracval(x) for x in confs["temperatures"].split()]
                if len(vals) < 1:
                    self.setting_error("Temperatures are incorrectly set.")
                else:
                    self.set_parameter("temperatures", vals)

    def _set_settings(self):
        self.set_settings()
        params = self._parameters

        # Supercell dimension for fc2
        if "dim_fc2" in params:
            self._settings.set_phonon_supercell_matrix(params["dim_fc2"])

        # Boundary mean free path for thermal conductivity calculation
        if "boundary_mfp" in params:
            self._settings.set_boundary_mfp(params["boundary_mfp"])

        # Calculate thermal conductivity in BTE-RTA
        if "bterta" in params:
            self._settings.set_is_bterta(params["bterta"])

        # Choice of thermal conductivity type
        if "conductivity_type" in params:
            self._settings.set_conductivity_type(params["conductivity_type"])

        # Solve collective phonons
        if "collective_phonon" in params:
            self._settings.set_solve_collective_phonon(params["collective_phonon"])

        # Compact force constants or full force constants
        if "compact_fc" in params:
            self._settings.set_is_compact_fc(params["compact_fc"])

        # Peierls type approximation for squared ph-ph interaction strength
        if "const_ave_pp" in params:
            self._settings.set_constant_averaged_pp_interaction(params["const_ave_pp"])

        # Trigger to create FORCES_FC2 and FORCES_FC3
        if "create_forces_fc2" in params:
            self._settings.set_create_forces_fc2(params["create_forces_fc2"])

        if "create_forces_fc3" in params:
            self._settings.set_create_forces_fc3(params["create_forces_fc3"])

        if "create_forces_fc3_file" in params:
            self._settings.set_create_forces_fc3_file(params["create_forces_fc3_file"])

        # Cutoff distance of third-order force constants. Elements where any
        # pair of atoms has larger distance than cut-off distance are set zero.
        if "cutoff_fc3_distance" in params:
            self._settings.set_cutoff_fc3_distance(params["cutoff_fc3_distance"])

        # Cutoff distance between pairs of displaced atoms used for supercell
        # creation with displacements and making third-order force constants
        if "cutoff_pair_distance" in params:
            self._settings.set_cutoff_pair_distance(params["cutoff_pair_distance"])

        # Emulate v1.x grid system and shortest vectors.
        if "emulate_v2" in params:
            self._settings.set_emulate_v2(params["emulate_v2"])

        # Grid addresses (sets of three integer values)
        if "grid_addresses" in params:
            self._settings.set_grid_addresses(params["grid_addresses"])

        # Grid points
        if "grid_points" in params:
            self._settings.set_grid_points(params["grid_points"])

        # Grid matrix
        if "grid_matrix" in params:
            self._settings.set_mesh_numbers(params["grid_matrix"])

        # Atoms are clamped under applied strain in Gruneisen parameter
        # calculation.
        if "ion_clamped" in params:
            self._settings.set_ion_clamped(params["ion_clamped"])

        # Take average in fc3-r2q transformation around three atoms
        if "fc3_r0_average" in params:
            self._settings.set_is_fc3_r0_average(params["fc3_r0_average"])

        # Calculate full ph-ph interaction strength for RTA conductivity
        if "full_pp" in params:
            self._settings.set_is_full_pp(params["full_pp"])

        # Calculate phonon-Gruneisen parameters
        if "gruneisen" in params:
            self._settings.set_is_gruneisen(params["gruneisen"])

        # Calculate imaginary part of self energy
        if "imag_self_energy" in params:
            self._settings.set_is_imag_self_energy(params["imag_self_energy"])

        # Calculate lifetime due to isotope scattering
        if "isotope" in params:
            self._settings.set_is_isotope(params["isotope"])

        # Calculate joint-DOS
        if "joint_dos" in params:
            self._settings.set_is_joint_dos(params["joint_dos"])

        # Sum partial kappa at q-stars
        if "kappa_star" in params:
            self._settings.set_is_kappa_star(params["kappa_star"])

        if "lapack_zheev_uplo" in params:
            self._settings.set_lapack_zheev_uplo(params["lapack_zheev_uplo"])

        # Calculate thermal conductivity in LBTE with Chaput's method
        if "lbte" in params:
            self._settings.set_is_lbte(params["lbte"])

        # Number of frequency points in a batch.
        if "num_points_in_batch" in params:
            self._settings.set_num_points_in_batch(params["num_points_in_batch"])

        # Calculate Normal and Umklapp processes
        if "N_U" in params:
            self._settings.set_is_N_U(params["N_U"])

        # Solve reducible collision matrix but not reduced matrix
        if "reducible_collision_matrix" in params:
            self._settings.set_is_reducible_collision_matrix(
                params["reducible_collision_matrix"]
            )

        # Symmetrize fc2 by index exchange
        if "symmetrize_fc2" in params:
            self._settings.set_is_symmetrize_fc2(params["symmetrize_fc2"])

        # Symmetrize phonon fc3 by index exchange
        if "symmetrize_fc3_q" in params:
            self._settings.set_is_symmetrize_fc3_q(params["symmetrize_fc3_q"])

        # Symmetrize fc3 by index exchange
        if "symmetrize_fc3_r" in params:
            self._settings.set_is_symmetrize_fc3_r(params["symmetrize_fc3_r"])

        # Mass variance parameters
        if "mass_variances" in params:
            self._settings.set_mass_variances(params["mass_variances"])

        # Maximum mean free path
        if "max_freepath" in params:
            self._settings.set_max_freepath(params["max_freepath"])

        # Output yaml filename instead of default filename of phono3py.yaml.
        if "output_yaml_filename" in params:
            self._settings.set_output_yaml_filename(params["output_yaml_filename"])

        # Cutoff frequency for pseudo inversion of collision matrix
        if "pinv_cutoff" in params:
            self._settings.set_pinv_cutoff(params["pinv_cutoff"])

        # Switch for pseudo-inverse method either taking abs or not.
        if "pinv_method" in params:
            self._settings.set_pinv_method(params["pinv_method"])

        # Switch for pseudo-inverse solver
        if "pinv_solver" in params:
            self._settings.set_pinv_solver(params["pinv_solver"])

        # Ph-ph interaction unit conversion factor
        if "pp_conversion_factor" in params:
            self._settings.set_pp_conversion_factor(params["pp_conversion_factor"])

        # Calculate real_self_energys
        if "real_self_energy" in params:
            self._settings.set_is_real_self_energy(params["real_self_energy"])

        # Read phonon-phonon interaction amplitudes from hdf5
        if "read_amplitude" in params:
            self._settings.set_read_amplitude(params["read_amplitude"])

        # Read collision matrix and gammas from hdf5
        if "read_collision" in params:
            self._settings.set_read_collision(params["read_collision"])

        # Read fc2 from hdf5
        if "read_fc2" in params:
            self._settings.set_read_fc2(params["read_fc2"])

        # Read fc3 from hdf5
        if "read_fc3" in params:
            self._settings.set_read_fc3(params["read_fc3"])

        # Read gammas from hdf5
        if "read_gamma" in params:
            self._settings.set_read_gamma(params["read_gamma"])

        # Read phonons from hdf5
        if "read_phonon" in params:
            self._settings.set_read_phonon(params["read_phonon"])

        # Read ph-ph interaction strength from hdf5
        if "read_pp" in params:
            self._settings.set_read_pp(params["read_pp"])

        # Scattering event class 1 or 2
        if "scattering_event_class" in params:
            self._settings.set_scattering_event_class(params["scattering_event_class"])

        # Cutoff width of smearing function (ratio to sigma value)
        if "sigma_cutoff_width" in params:
            self._settings.set_sigma_cutoff_width(params["sigma_cutoff_width"])

        # Calculate spectral_functions
        if "spectral_function" in params:
            self._settings.set_is_spectral_function(params["spectral_function"])

        # Subtract residual forces to create FORCES_FC2 and FORCES_FC3
        if "subtract_forces" in params:
            self._settings.set_subtract_forces(params["subtract_forces"])

        # Temperatures for scatterings
        if "temperatures" in params:
            self._settings.set_temperatures(params["temperatures"])

        # Use averaged ph-ph interaction
        if "use_ave_pp" in params:
            self._settings.set_use_ave_pp(params["use_ave_pp"])

        # Use generalized regular grid
        if "use_grg" in params:
            self._settings.set_use_grg(params["use_grg"])

        # Write detailed imag-part of self energy to hdf5
        if "write_gamma_detail" in params:
            self._settings.set_write_gamma_detail(params["write_gamma_detail"])

        # Write imag-part of self energy to hdf5
        if "write_gamma" in params:
            self._settings.set_write_gamma(params["write_gamma"])

        # Write collision matrix and gammas to hdf5
        if "write_collision" in params:
            self._settings.set_write_collision(params["write_collision"])

        # Write all phonons on grid points to hdf5
        if "write_phonon" in params:
            self._settings.set_write_phonon(params["write_phonon"])

        # Write phonon-phonon interaction amplitudes to hdf5
        if "write_pp" in params:
            self._settings.set_write_pp(params["write_pp"])

        # Write direct solution of LBTE to hdf5 files
        if "write_LBTE_solution" in params:
            self._settings.set_write_LBTE_solution(params["write_LBTE_solution"])



"""Example to run thermal conductivity of Si."""
import numpy as np
from phonopy.interface.vasp import read_vasp

import phono3py
from phono3py import Phono3py


def run_thermal_conductivity():
    """Run RTA thermal conductivity calculation from input files."""
    ph3 = phono3py.load("phono3py_disp.yaml", log_level=2)
    ph3.mesh_numbers = [11, 11, 11]
    ph3.init_phph_interaction()
    ph3.run_thermal_conductivity(temperatures=range(0, 1001, 10), write_kappa=True)
    # Conductivity_RTA object
    print(ph3.thermal_conductivity.kappa)


def create_supercells_with_displacements():
    """Create supercells with displacements."""
    cell = read_vasp("POSCAR-unitcell")
    ph3 = Phono3py(cell, np.diag([2, 2, 2]), primitive_matrix="F")
    ph3.generate_displacements(distance=0.03)
    print(ph3.supercells_with_displacements)  # List of PhonopyAtoms
    print(ph3.displacements.shape)  # (supercells, atoms, xyz)


if __name__ == "__main__":
    create_supercells_with_displacements()
    run_thermal_conductivity()



"""Tests for grids."""
from __future__ import annotations

import numpy as np
import pytest
from phonopy import Phonopy
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.symmetry import Symmetry

from phono3py import Phono3py
from phono3py.other.tetrahedron_method import get_tetrahedra_relative_grid_address
from phono3py.phonon.grid import (
    BZGrid,
    GridMatrix,
    _can_use_std_lattice,
    _get_grid_points_by_bz_rotations_c,
    _get_grid_points_by_bz_rotations_py,
    _get_grid_points_by_rotations,
    _relocate_BZ_grid_address,
    get_grid_point_from_address,
    get_grid_point_from_address_py,
    get_ir_grid_points,
)


def _get_qpoints(adrs, bzgrid):
    return np.dot(
        (adrs * 2 + bzgrid.PS) / bzgrid.D_diag.astype("double") / 2, bzgrid.Q.T
    )


def test_get_grid_point_from_address(agno2_cell):
    """Test for get_grid_point_from_address.

    Compare get_grid_point_from_address from spglib and that
    written in python with mesh numbers.

    """
    mesh = (10, 10, 10)

    for address in list(np.ndindex(mesh)):
        gp_spglib = get_grid_point_from_address(address, mesh)
        gp_py = get_grid_point_from_address_py(address, mesh)
        # print("%s %d %d" % (address, gp_spglib, gp_py))
        np.testing.assert_equal(gp_spglib, gp_py)


def test_BZGrid(si_pbesol_111: Phono3py):
    """Tests of BZGrid type1 and type2."""
    lat = si_pbesol_111.primitive.cell
    reclat = np.linalg.inv(lat)
    mesh = [4, 4, 4]

    gp_map2 = [
        0,
        1,
        2,
        4,
        5,
        6,
        7,
        8,
        9,
        11,
        12,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        34,
        35,
        36,
        40,
        41,
        43,
        44,
        46,
        47,
        48,
        49,
        50,
        54,
        56,
        57,
        59,
        60,
        61,
        65,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        77,
        78,
        79,
        83,
        84,
        85,
        86,
        87,
        88,
        89,
    ]

    bzgrid1 = BZGrid(mesh, lattice=lat, store_dense_gp_map=False)
    bzgrid2 = BZGrid(mesh, lattice=lat, store_dense_gp_map=True)

    adrs1 = bzgrid1.addresses[: np.prod(mesh)]
    adrs2 = bzgrid2.addresses[bzgrid2.gp_map[:-1]]
    assert ((adrs1 - adrs2) % mesh == 0).all()
    np.testing.assert_equal(bzgrid1.addresses.shape, bzgrid2.addresses.shape)
    # print("".join(["%d, " % i for i in bzgrid2.gp_map.ravel()]))
    np.testing.assert_equal(bzgrid2.gp_map.ravel(), gp_map2)

    dist1 = np.sqrt((np.dot(adrs1, reclat.T) ** 2).sum(axis=1))
    dist2 = np.sqrt((np.dot(adrs2, reclat.T) ** 2).sum(axis=1))
    np.testing.assert_allclose(dist1, dist2, atol=1e-8)


def test_BZGrid_bzg2grg(si_pbesol_111):
    """Test of mapping of BZGrid to GRGrid.

    This mapping table is stored in BZGrid, but also determined by
    get_grid_point_from_address. This test checks the consistency.

    """
    lat = si_pbesol_111.primitive.cell
    mesh = [4, 4, 4]
    bzgrid1 = BZGrid(mesh, lattice=lat, store_dense_gp_map=False)
    grg = []
    for i in range(len(bzgrid1.addresses)):
        grg.append(get_grid_point_from_address(bzgrid1.addresses[i], mesh))
    np.testing.assert_equal(grg, bzgrid1.bzg2grg)

    bzgrid2 = BZGrid(mesh, lattice=lat, store_dense_gp_map=True)
    grg = []
    for i in range(len(bzgrid2.addresses)):
        grg.append(get_grid_point_from_address(bzgrid2.addresses[i], mesh))
    np.testing.assert_equal(grg, bzgrid2.bzg2grg)


def test_BZGrid_SNF(si_pbesol_111: Phono3py):
    """Test of SNF in BZGrid."""
    lat = si_pbesol_111.primitive.cell
    mesh = 10
    bzgrid1 = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=si_pbesol_111.primitive_symmetry.dataset,
        use_grg=True,
        store_dense_gp_map=False,
    )
    _test_BZGrid_SNF(bzgrid1)

    bzgrid2 = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=si_pbesol_111.primitive_symmetry.dataset,
        use_grg=True,
        store_dense_gp_map=True,
    )
    _test_BZGrid_SNF(bzgrid2)


def test_BZGrid_SNF_with_tmat(si_pbesol_111: Phono3py):
    """Test of SNF in BZGrid with transformation matrix."""
    lat = si_pbesol_111.primitive.cell
    mesh = 10
    tmat = [[0.0, 0.5, 0.5], [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]]
    bzgrid1 = BZGrid(
        mesh,
        lattice=lat,
        transformation_matrix=tmat,
        use_grg=True,
    )
    _test_BZGrid_SNF(bzgrid1)


def test_BZGrid_SNF_with_no_tmat_and_dataset(si_pbesol_111: Phono3py):
    """Test of SNF in BZGrid with transformation matrix."""
    lat = si_pbesol_111.primitive.cell
    mesh = 10
    with pytest.raises(RuntimeError):
        BZGrid(
            mesh,
            lattice=lat,
            use_grg=True,
        )


def test_BZGrid_SNF_with_negative_tmat(si_pbesol_111: Phono3py):
    """Test of SNF in BZGrid with negative transformation matrix."""
    lat = si_pbesol_111.primitive.cell
    mesh = 10
    tmat = [[0.0, -0.5, -0.5], [-0.5, 0.0, -0.5], [-0.5, -0.5, 0.0]]
    with pytest.raises(RuntimeError):
        BZGrid(
            mesh,
            lattice=lat,
            transformation_matrix=tmat,
            use_grg=True,
        )


def test_BZGrid_SNF_with_non_integer_inv_tmat(si_pbesol_111: Phono3py):
    """Test of SNF in BZGrid with negative transformation matrix."""
    lat = si_pbesol_111.primitive.cell
    mesh = 10
    tmat = [[0, 1, 1], [1, 0, 1], [1, 1, 0]]
    with pytest.raises(RuntimeError):
        BZGrid(
            mesh,
            lattice=lat,
            transformation_matrix=tmat,
            use_grg=True,
        )


def _test_BZGrid_SNF(bzgrid: BZGrid):
    # from phonopy.structure.atoms import PhonopyAtoms
    # from phonopy.interface.vasp import get_vasp_structure_lines

    A = bzgrid.grid_matrix
    D_diag = bzgrid.D_diag
    P = bzgrid.P
    Q = bzgrid.Q
    np.testing.assert_equal(np.dot(P, np.dot(A, Q)), np.diag(D_diag))

    # print(D_diag)
    # grg2bzg = bzgrid.grg2bzg
    # qpoints = np.dot(bzgrid.addresses[grg2bzg], bzgrid.QDinv.T)
    # cell = PhonopyAtoms(cell=np.linalg.inv(lat).T,
    #                     scaled_positions=qpoints,
    #                     numbers=[1,] * len(qpoints))
    # print("\n".join(get_vasp_structure_lines(cell)))

    gr_addresses = bzgrid.addresses[bzgrid.grg2bzg]
    # print(D_diag)
    # print(len(gr_addresses))
    # for line in gr_addresses.reshape(-1, 12):
    #     print("".join(["%d, " % i for i in line]))

    ref = [
        0,
        0,
        0,
        -1,
        0,
        0,
        0,
        1,
        0,
        1,
        1,
        0,
        0,
        -2,
        0,
        -1,
        -2,
        0,
        0,
        -1,
        0,
        -1,
        -1,
        0,
        0,
        0,
        1,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        0,
        2,
        1,
        1,
        2,
        1,
        0,
        -1,
        1,
        -1,
        -1,
        1,
        0,
        0,
        -2,
        -1,
        0,
        -2,
        0,
        1,
        2,
        1,
        1,
        2,
        0,
        -2,
        -2,
        -1,
        -2,
        -2,
        0,
        -1,
        -2,
        -1,
        -1,
        -2,
        0,
        0,
        -1,
        -1,
        0,
        -1,
        0,
        1,
        -1,
        -1,
        1,
        -1,
        0,
        -2,
        -1,
        -1,
        -2,
        -1,
        0,
        -1,
        -1,
        -1,
        -1,
        -1,
    ]

    assert (
        ((np.reshape(ref, (-1, 3)) - gr_addresses) % bzgrid.D_diag).ravel() == 0
    ).all()

    if bzgrid.symmetry_dataset is not None:
        ref_rots = [
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            2,
            1,
            -1,
            2,
            2,
            -1,
            1,
            0,
            0,
            2,
            -1,
            0,
            4,
            0,
            -1,
            1,
            0,
            0,
            0,
            -1,
            1,
            2,
            -2,
            1,
            -1,
            0,
            0,
            0,
            -1,
            0,
            -2,
            -2,
            1,
            -1,
            0,
            0,
            0,
            1,
            -1,
            0,
            0,
            -1,
            -1,
            0,
            0,
            -2,
            1,
            0,
            -2,
            2,
            -1,
            -1,
            0,
            0,
            -2,
            -1,
            1,
            -4,
            0,
            1,
            -1,
            -1,
            1,
            0,
            -1,
            1,
            -2,
            -1,
            2,
            -1,
            -1,
            1,
            0,
            -2,
            1,
            0,
            -3,
            2,
            -1,
            -1,
            1,
            -2,
            -1,
            1,
            -2,
            -3,
            2,
            -1,
            -1,
            1,
            -2,
            0,
            1,
            -4,
            -1,
            2,
            1,
            1,
            -1,
            0,
            1,
            -1,
            0,
            3,
            -2,
            1,
            1,
            -1,
            2,
            0,
            -1,
            2,
            1,
            -2,
            1,
            1,
            -1,
            2,
            1,
            -1,
            4,
            1,
            -2,
            1,
            1,
            -1,
            0,
            2,
            -1,
            2,
            3,
            -2,
            -1,
            1,
            0,
            -2,
            0,
            1,
            -2,
            1,
            1,
            -1,
            1,
            0,
            -2,
            1,
            0,
            -4,
            1,
            1,
            -1,
            1,
            0,
            0,
            2,
            -1,
            -2,
            3,
            -1,
            -1,
            1,
            0,
            0,
            1,
            0,
            0,
            3,
            -1,
            1,
            -1,
            0,
            2,
            0,
            -1,
            4,
            -1,
            -1,
            1,
            -1,
            0,
            0,
            -1,
            0,
            2,
            -1,
            -1,
            1,
            -1,
            0,
            0,
            -2,
            1,
            0,
            -3,
            1,
            1,
            -1,
            0,
            2,
            -1,
            0,
            2,
            -3,
            1,
            -1,
            0,
            0,
            0,
            -1,
            0,
            0,
            0,
            -1,
            -1,
            0,
            0,
            -2,
            -1,
            1,
            -2,
            -2,
            1,
            -1,
            0,
            0,
            -2,
            1,
            0,
            -4,
            0,
            1,
            -1,
            0,
            0,
            0,
            1,
            -1,
            -2,
            2,
            -1,
            1,
            0,
            0,
            0,
            1,
            0,
            2,
            2,
            -1,
            1,
            0,
            0,
            0,
            -1,
            1,
            0,
            0,
            1,
            1,
            0,
            0,
            2,
            -1,
            0,
            2,
            -2,
            1,
            1,
            0,
            0,
            2,
            1,
            -1,
            4,
            0,
            -1,
            1,
            1,
            -1,
            0,
            1,
            -1,
            2,
            1,
            -2,
            1,
            1,
            -1,
            0,
            2,
            -1,
            0,
            3,
            -2,
            1,
            1,
            -1,
            2,
            1,
            -1,
            2,
            3,
            -2,
            1,
            1,
            -1,
            2,
            0,
            -1,
            4,
            1,
            -2,
            -1,
            -1,
            1,
            0,
            -1,
            1,
            0,
            -3,
            2,
            -1,
            -1,
            1,
            -2,
            0,
            1,
            -2,
            -1,
            2,
            -1,
            -1,
            1,
            -2,
            -1,
            1,
            -4,
            -1,
            2,
            -1,
            -1,
            1,
            0,
            -2,
            1,
            -2,
            -3,
            2,
            1,
            -1,
            0,
            2,
            0,
            -1,
            2,
            -1,
            -1,
            1,
            -1,
            0,
            2,
            -1,
            0,
            4,
            -1,
            -1,
            1,
            -1,
            0,
            0,
            -2,
            1,
            2,
            -3,
            1,
            1,
            -1,
            0,
            0,
            -1,
            0,
            0,
            -3,
            1,
            -1,
            1,
            0,
            -2,
            0,
            1,
            -4,
            1,
            1,
            -1,
            1,
            0,
            0,
            1,
            0,
            -2,
            1,
            1,
            -1,
            1,
            0,
            0,
            2,
            -1,
            0,
            3,
            -1,
            -1,
            1,
            0,
            -2,
            1,
            0,
            -2,
            3,
            -1,
        ]

        np.testing.assert_equal(ref_rots, bzgrid.rotations.ravel())


def test_BZGrid_SNF_hexagonal(aln_lda):
    """Test of SNF in BZGrid."""
    lat = aln_lda.primitive.cell
    mesh = 20
    bzgrid = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=aln_lda.primitive_symmetry.dataset,
    )
    np.testing.assert_equal(bzgrid.D_diag, [7, 7, 4])

    bzgrid = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=aln_lda.primitive_symmetry.dataset,
        use_grg=True,
    )
    np.testing.assert_equal(bzgrid.D_diag, [7, 7, 4])

    bzgrid = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=aln_lda.primitive_symmetry.dataset,
        use_grg=True,
        force_SNF=True,
    )
    np.testing.assert_equal(bzgrid.D_diag, [1, 7, 28])

    bzgrid = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=aln_lda.primitive_symmetry.dataset,
        use_grg=True,
        force_SNF=True,
        SNF_coordinates="direct",
    )
    np.testing.assert_equal(bzgrid.D_diag, [1, 7, 28])


def test_BZGrid_SNF_nonprimitive(si_pbesol_111):
    """Test of SNF in BZGrid."""
    lat = si_pbesol_111.supercell.cell
    mesh = 20
    with pytest.warns(
        RuntimeWarning, match="Non primitive cell input. Unable to use GR-grid."
    ):
        bzgrid = BZGrid(
            mesh,
            lattice=lat,
            symmetry_dataset=si_pbesol_111.symmetry.dataset,
            use_grg=True,
        )
    np.testing.assert_equal(bzgrid.D_diag, [4, 4, 4])
    identity = np.eye(3, dtype=int)
    np.testing.assert_equal(bzgrid.P, identity)
    np.testing.assert_equal(bzgrid.Q, identity)


def test_SNF_tetrahedra_relative_grid(aln_lda):
    """Test relative grid addresses under GR-grid.

    Under GR-grid, grid point addressing becomes different from ordinal uniform
    grid. But P and Q matrices can be used to map betweewn these grid systems.
    In this test, the agreement is checked by representing them in Cartesian
    coordinates.

    """
    lat = aln_lda.primitive.cell
    mesh = 25

    for snf_coordinates, d_diag in zip(
        ("direct", "reciprocal"), ([1, 9, 45], [1, 9, 45])
    ):
        bzgrid = BZGrid(
            mesh,
            lattice=lat,
            symmetry_dataset=aln_lda.primitive_symmetry.dataset,
            use_grg=True,
            force_SNF=True,
            SNF_coordinates=snf_coordinates,
        )

        np.testing.assert_equal(bzgrid.D_diag, d_diag)

        plat = np.linalg.inv(aln_lda.primitive.cell)
        mlat = bzgrid.microzone_lattice
        tetrahedra = get_tetrahedra_relative_grid_address(mlat)
        snf_tetrahedra = np.dot(tetrahedra, bzgrid.P.T)

        for mtet, ptet in zip(tetrahedra, snf_tetrahedra):
            np.testing.assert_allclose(
                np.dot(mtet, mlat.T),
                np.dot(np.dot(ptet, bzgrid.QDinv.T), plat.T),
                atol=1e-8,
            )


def test_get_grid_points_by_bz_rotations(si_pbesol_111):
    """Rotate grid point by rotations with and without considering BZ surface.

    The following three methods are tested between type-1 and type-2.

        _get_grid_points_by_rotations
        _get_grid_points_by_bz_rotations_c
        _get_grid_points_by_bz_rotations_py

    """
    ref10_type1 = [
        10,
        26,
        10,
        26,
        26,
        10,
        26,
        10,
        88,
        80,
        200,
        208,
        200,
        208,
        88,
        80,
        208,
        88,
        80,
        200,
        208,
        88,
        80,
        200,
        26,
        10,
        26,
        10,
        10,
        26,
        10,
        26,
        200,
        208,
        88,
        80,
        88,
        80,
        200,
        208,
        80,
        200,
        208,
        88,
        80,
        200,
        208,
        88,
    ]
    ref12_type2 = [
        12,
        39,
        12,
        39,
        39,
        12,
        39,
        12,
        122,
        109,
        265,
        278,
        265,
        278,
        122,
        109,
        278,
        122,
        109,
        265,
        278,
        122,
        109,
        265,
        39,
        12,
        39,
        12,
        12,
        39,
        12,
        39,
        265,
        278,
        122,
        109,
        122,
        109,
        265,
        278,
        109,
        265,
        278,
        122,
        109,
        265,
        278,
        122,
    ]

    ref10_bz_type1 = [
        10,
        26,
        260,
        270,
        269,
        258,
        271,
        259,
        88,
        285,
        200,
        328,
        322,
        208,
        291,
        286,
        327,
        292,
        287,
        321,
        326,
        290,
        80,
        323,
        269,
        258,
        271,
        259,
        10,
        26,
        260,
        270,
        200,
        328,
        88,
        285,
        291,
        286,
        322,
        208,
        80,
        323,
        326,
        290,
        287,
        321,
        327,
        292,
    ]
    ref12_bz_type2 = [
        12,
        39,
        15,
        41,
        40,
        13,
        42,
        14,
        122,
        110,
        265,
        281,
        267,
        278,
        124,
        111,
        280,
        125,
        112,
        266,
        279,
        123,
        109,
        268,
        40,
        13,
        42,
        14,
        12,
        39,
        15,
        41,
        265,
        281,
        122,
        110,
        124,
        111,
        267,
        278,
        109,
        268,
        279,
        123,
        112,
        266,
        280,
        125,
    ]

    lat = si_pbesol_111.primitive.cell
    mesh = 20

    bz_grid_type1 = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=si_pbesol_111.primitive_symmetry.dataset,
        use_grg=True,
        store_dense_gp_map=False,
    )
    bz_grid_type2 = BZGrid(
        mesh,
        lattice=lat,
        symmetry_dataset=si_pbesol_111.primitive_symmetry.dataset,
        use_grg=True,
        store_dense_gp_map=True,
    )

    # Check data consistency by reducing to GR-grid.
    # Grid point 10 in type-1 and 12 in type-2 are the same points in GR-grid.
    assert bz_grid_type1.bzg2grg[10] == bz_grid_type2.bzg2grg[12]
    np.testing.assert_equal(
        bz_grid_type1.bzg2grg[ref10_type1], bz_grid_type2.bzg2grg[ref12_type2]
    )
    np.testing.assert_equal(
        bz_grid_type1.bzg2grg[ref10_type1], bz_grid_type1.bzg2grg[ref10_bz_type1]
    )
    np.testing.assert_equal(
        bz_grid_type1.bzg2grg[ref10_type1], bz_grid_type2.bzg2grg[ref12_bz_type2]
    )

    bzgps = _get_grid_points_by_rotations(10, bz_grid_type1, bz_grid_type1.rotations)
    np.testing.assert_equal(bzgps, ref10_type1)

    bzgps = _get_grid_points_by_rotations(12, bz_grid_type2, bz_grid_type2.rotations)
    np.testing.assert_equal(bzgps, ref12_type2)

    bzgps = _get_grid_points_by_bz_rotations_c(
        10, bz_grid_type1, bz_grid_type1.rotations
    )
    np.testing.assert_equal(bzgps, ref10_bz_type1)

    bzgps = _get_grid_points_by_bz_rotations_c(
        12, bz_grid_type2, bz_grid_type2.rotations
    )
    np.testing.assert_equal(bzgps, ref12_bz_type2)

    bzgps = _get_grid_points_by_bz_rotations_py(
        10, bz_grid_type1, bz_grid_type1.rotations
    )
    np.testing.assert_equal(bzgps, ref10_bz_type1)

    bzgps = _get_grid_points_by_bz_rotations_py(
        12, bz_grid_type2, bz_grid_type2.rotations
    )
    np.testing.assert_equal(bzgps, ref12_bz_type2)

    # Exhaustive consistency check among methods
    for grgp in range(np.prod(bz_grid_type1.D_diag)):
        bzgp_type1 = bz_grid_type1.grg2bzg[grgp]
        bzgp_type2 = bz_grid_type2.grg2bzg[grgp]

        rot_grgps = bz_grid_type1.bzg2grg[
            _get_grid_points_by_rotations(
                bzgp_type1, bz_grid_type1, bz_grid_type1.rotations
            )
        ]

        np.testing.assert_equal(
            rot_grgps,
            bz_grid_type2.bzg2grg[
                _get_grid_points_by_rotations(
                    bzgp_type2, bz_grid_type2, bz_grid_type2.rotations
                )
            ],
        )

        np.testing.assert_equal(
            _get_grid_points_by_bz_rotations_c(
                bzgp_type1, bz_grid_type1, bz_grid_type1.rotations
            ),
            _get_grid_points_by_bz_rotations_py(
                bzgp_type1, bz_grid_type1, bz_grid_type1.rotations
            ),
        )

        np.testing.assert_equal(
            _get_grid_points_by_bz_rotations_c(
                bzgp_type2, bz_grid_type2, bz_grid_type2.rotations
            ),
            _get_grid_points_by_bz_rotations_py(
                bzgp_type2, bz_grid_type2, bz_grid_type2.rotations
            ),
        )

        np.testing.assert_equal(
            rot_grgps,
            bz_grid_type1.bzg2grg[
                _get_grid_points_by_bz_rotations_c(
                    bzgp_type1, bz_grid_type1, bz_grid_type1.rotations
                )
            ],
        )

        np.testing.assert_equal(
            rot_grgps,
            bz_grid_type2.bzg2grg[
                _get_grid_points_by_bz_rotations_c(
                    bzgp_type2, bz_grid_type2, bz_grid_type2.rotations
                )
            ],
        )

    # for gps in bzgps.reshape(-1, 12):
    #     print("".join(["%d, " % gp for gp in gps]))


def test_can_use_std_lattice():
    """Test of can_use_std_lattice."""
    conv_lat = [[6.06531185, 0.0, 0.0], [0.0, 0.0, 6.06531185], [0.0, -6.06531185, 0.0]]
    std_lattice = [
        [6.06531185, 0.0, 0.0],
        [0.0, 6.06531185, 0.0],
        [0.0, 0.0, 6.06531185],
    ]
    tmat = [[0.0, 0.5, 0.5], [-0.5, -0.5, 0.0], [0.5, 0.0, 0.5]]
    rotations = [
        [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        [[0, 1, 0], [0, 0, 1], [-1, -1, -1]],
        [[0, 0, 1], [-1, -1, -1], [1, 0, 0]],
        [[-1, -1, -1], [1, 0, 0], [0, 1, 0]],
        [[-1, -1, -1], [0, 0, 1], [0, 1, 0]],
        [[1, 0, 0], [-1, -1, -1], [0, 0, 1]],
        [[0, 1, 0], [1, 0, 0], [-1, -1, -1]],
        [[0, 0, 1], [0, 1, 0], [1, 0, 0]],
        [[-1, -1, -1], [1, 0, 0], [0, 0, 1]],
        [[1, 0, 0], [0, 1, 0], [-1, -1, -1]],
        [[0, 1, 0], [0, 0, 1], [1, 0, 0]],
        [[0, 0, 1], [-1, -1, -1], [0, 1, 0]],
        [[1, 0, 0], [-1, -1, -1], [0, 1, 0]],
        [[0, 1, 0], [1, 0, 0], [0, 0, 1]],
        [[0, 0, 1], [0, 1, 0], [-1, -1, -1]],
        [[-1, -1, -1], [0, 0, 1], [1, 0, 0]],
        [[0, 1, 0], [-1, -1, -1], [0, 0, 1]],
        [[0, 0, 1], [1, 0, 0], [-1, -1, -1]],
        [[-1, -1, -1], [0, 1, 0], [1, 0, 0]],
        [[1, 0, 0], [0, 0, 1], [0, 1, 0]],
        [[0, 0, 1], [1, 0, 0], [0, 1, 0]],
        [[-1, -1, -1], [0, 1, 0], [0, 0, 1]],
        [[1, 0, 0], [0, 0, 1], [-1, -1, -1]],
        [[0, 1, 0], [-1, -1, -1], [1, 0, 0]],
    ]

    assert _can_use_std_lattice(conv_lat, tmat, std_lattice, rotations)


def test_aln_BZGrid_with_shift(aln_cell: PhonopyAtoms):
    """Test BZGrid with shift using AlN."""
    mesh = [5, 5, 4]
    symmetry = Symmetry(aln_cell)

    # Without shift
    bzgrid = BZGrid(mesh, lattice=aln_cell.cell, symmetry_dataset=symmetry.dataset)
    ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bzgrid)
    np.testing.assert_equal(
        ir_grid_points, [0, 1, 2, 6, 7, 25, 26, 27, 31, 32, 50, 51, 52, 56, 57]
    )
    np.testing.assert_equal(
        ir_grid_weights, [1, 6, 6, 6, 6, 2, 12, 12, 12, 12, 1, 6, 6, 6, 6]
    )

    # With shift
    bzgrid = BZGrid(
        mesh,
        lattice=aln_cell.cell,
        symmetry_dataset=symmetry.dataset,
        is_shift=[False, False, True],
    )
    ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bzgrid)
    np.testing.assert_equal(ir_grid_points, [0, 1, 2, 6, 7, 25, 26, 27, 31, 32])
    np.testing.assert_equal(ir_grid_weights, [2, 12, 12, 12, 12, 2, 12, 12, 12, 12])

    q_from_phonopy = [
        [0.0000000, 0.0000000, 0.1250000],
        [0.2000000, 0.0000000, 0.1250000],
        [0.4000000, 0.0000000, 0.1250000],
        [0.2000000, 0.2000000, 0.1250000],
        [-0.6000000, 0.2000000, 0.1250000],
        [0.0000000, 0.0000000, 0.3750000],
        [0.2000000, 0.0000000, 0.3750000],
        [0.4000000, 0.0000000, 0.3750000],
        [0.2000000, 0.2000000, 0.3750000],
        [-0.6000000, 0.2000000, 0.3750000],
    ]

    for adrs, q_phonopy in zip(
        bzgrid.addresses[bzgrid.grg2bzg[ir_grid_points]], q_from_phonopy
    ):
        q = np.dot(
            bzgrid.Q, (adrs * 2 + bzgrid.PS) / bzgrid.D_diag.astype("double") / 2
        )
        diff = q - q_phonopy
        diff -= np.rint(diff)
        np.testing.assert_allclose(diff, [0, 0, 0])
        q_phonopy_norm = np.linalg.norm(np.dot(np.linalg.inv(aln_cell.cell), q_phonopy))
        q_norm = np.linalg.norm(np.dot(np.linalg.inv(aln_cell.cell), q))
        np.testing.assert_almost_equal(q_phonopy_norm, q_norm)


@pytest.mark.parametrize(
    "is_shift",
    [
        [True, False, False],
        [False, True, False],
        [True, True, False],
        [True, False, True],
        [False, True, True],
        [True, True, True],
    ],
)
def test_aln_BZGrid_with_shift_broken_symmetry(aln_cell: PhonopyAtoms, is_shift: list):
    """Test broken symmetry of BZGrid with shift using AlN."""
    mesh = [5, 5, 4]
    symmetry = Symmetry(aln_cell)

    with pytest.raises(RuntimeError):
        BZGrid(
            mesh,
            lattice=aln_cell.cell,
            symmetry_dataset=symmetry.dataset,
            is_shift=is_shift,
        )


def test_agno2_BZGrid_with_shift(agno2_cell: PhonopyAtoms):
    """Test BZGrid with shift using AgNO2."""
    mesh = 15
    ph = Phonopy(agno2_cell, supercell_matrix=[1, 1, 1], primitive_matrix="auto")

    # from phonopy.interface.vasp import get_vasp_structure_lines
    # print("\n".join(get_vasp_structure_lines(ph.primitive)))

    # Without shift
    bzgrid = BZGrid(
        mesh,
        lattice=ph.primitive.cell,
        symmetry_dataset=ph.primitive_symmetry.dataset,
        use_grg=True,
    )
    ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bzgrid)
    np.testing.assert_equal(bzgrid.grid_matrix, [[0, 5, 5], [2, 0, 2], [3, 3, 0]])
    np.testing.assert_equal(
        ir_grid_points, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 14, 15, 18, 20, 24, 30]
    )
    np.testing.assert_equal(
        ir_grid_weights, [1, 8, 4, 4, 4, 4, 2, 8, 4, 4, 2, 2, 4, 2, 2, 2, 2, 1]
    )
    bzgrid_no_shift = bzgrid

    # Digonal elements represent orthorhombic microzone.
    shift_ref = np.diagonal(bzgrid.microzone_lattice) / 2

    # With shift +c
    bzgrid = BZGrid(
        mesh,
        lattice=ph.primitive.cell,
        symmetry_dataset=ph.primitive_symmetry.dataset,
        use_grg=True,
        is_shift=[False, False, True],
    )
    ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bzgrid)

    np.testing.assert_equal(
        ir_grid_points, [0, 1, 2, 3, 4, 5, 11, 12, 14, 15, 17, 18, 20, 24, 30]
    )
    np.testing.assert_equal(
        ir_grid_weights, [2, 8, 4, 8, 4, 2, 4, 4, 4, 4, 4, 4, 2, 4, 2]
    )

    q_shift_c = _get_qpoints(bzgrid.addresses[bzgrid.grg2bzg], bzgrid)
    q_noshift = _get_qpoints(
        bzgrid_no_shift.addresses[bzgrid_no_shift.grg2bzg], bzgrid_no_shift
    )
    diff = q_shift_c - q_noshift
    diff -= np.rint(diff)
    diff_cart = np.dot(diff, np.linalg.inv(ph.primitive.cell).T)
    np.testing.assert_allclose(diff_cart - [0, 0, 3.33538325e-02], 0, atol=1e-8)
    np.testing.assert_allclose(diff_cart[0][2], shift_ref[2], atol=1e-8)

    # With shift +a, +b
    bzgrid = BZGrid(
        mesh,
        lattice=ph.primitive.cell,
        symmetry_dataset=ph.primitive_symmetry.dataset,
        use_grg=True,
        is_shift=[True, True, False],
    )
    ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bzgrid)

    np.testing.assert_equal(ir_grid_points, [0, 1, 2, 3, 4, 5, 6, 8, 9, 12])
    np.testing.assert_equal(ir_grid_weights, [4, 8, 8, 4, 8, 8, 4, 8, 4, 4])

    q_shift_ab = _get_qpoints(bzgrid.addresses[bzgrid.grg2bzg], bzgrid)
    diff = q_shift_ab - q_noshift
    diff -= np.rint(diff)
    diff_cart = np.dot(diff, np.linalg.inv(ph.primitive.cell).T)
    np.testing.assert_allclose(
        diff_cart - [3.03777935e-02, 3.89622460e-02, 0], 0, atol=1e-8
    )
    np.testing.assert_allclose(diff_cart[0][[0, 1]], shift_ref[[0, 1]], atol=1e-8)

    # With shift +a, +b, +c
    bzgrid = BZGrid(
        mesh,
        lattice=ph.primitive.cell,
        symmetry_dataset=ph.primitive_symmetry.dataset,
        use_grg=True,
        is_shift=[True, True, True],
    )
    ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bzgrid)

    np.testing.assert_equal(ir_grid_points, [0, 1, 2, 3, 5, 7, 8, 9])
    np.testing.assert_equal(ir_grid_weights, [8, 8, 4, 8, 8, 8, 8, 8])

    q_shift_ab = _get_qpoints(bzgrid.addresses[bzgrid.grg2bzg], bzgrid)
    diff = q_shift_ab - q_noshift
    diff -= np.rint(diff)
    diff_cart = np.dot(diff, np.linalg.inv(ph.primitive.cell).T)
    np.testing.assert_allclose(
        diff_cart - [0.03037779, 0.03896225, 0.03335383], 0, atol=1e-8
    )
    np.testing.assert_allclose(diff_cart[0], shift_ref, atol=1e-8)


def test_relocate_BZ_grid_address_FCC():
    """Test of _relocate_BZ_grid_address by FCC."""
    D_diag = [4, 4, 4]
    Q = np.eye(3, dtype="int_")
    reciprocal_lattice = np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]], dtype="double")
    bz_grid_addresses, bz_map, bzg2grg = _relocate_BZ_grid_address(
        D_diag,
        Q,
        reciprocal_lattice,  # column vectors
        PS=None,
        store_dense_gp_map=True,
    )
    ref_bz_grid_addresses = [
        [0, 0, 0],
        [1, 0, 0],
        [-2, 0, 0],
        [2, 0, 0],
        [-1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [2, 1, 0],
        [-1, 1, 0],
        [0, -2, 0],
        [0, 2, 0],
        [1, 2, 0],
        [-2, -2, 0],
        [2, 2, 0],
        [-1, -2, 0],
        [0, -1, 0],
        [1, -1, 0],
        [-2, -1, 0],
        [-1, -1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [2, 0, 1],
        [-1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
        [2, 1, 1],
        [-1, 1, 1],
        [0, 2, 1],
        [1, 2, 1],
        [2, 2, 1],
        [-1, -2, 1],
        [-1, -2, -3],
        [-1, 2, 1],
        [3, 2, 1],
        [0, -1, 1],
        [1, -1, 1],
        [-2, -1, 1],
        [-2, -1, -3],
        [2, -1, 1],
        [2, 3, 1],
        [-1, -1, 1],
        [0, 0, -2],
        [0, 0, 2],
        [1, 0, 2],
        [-2, 0, -2],
        [2, 0, 2],
        [-1, 0, -2],
        [0, 1, 2],
        [1, 1, 2],
        [2, 1, 2],
        [-1, 1, -2],
        [-1, 1, 2],
        [-1, -3, -2],
        [3, 1, 2],
        [0, -2, -2],
        [0, 2, 2],
        [1, 2, 2],
        [-2, -2, -2],
        [2, 2, 2],
        [-1, -2, -2],
        [0, -1, -2],
        [1, -1, -2],
        [1, -1, 2],
        [1, 3, 2],
        [-3, -1, -2],
        [-2, -1, -2],
        [-1, -1, -2],
        [0, 0, -1],
        [1, 0, -1],
        [-2, 0, -1],
        [-1, 0, -1],
        [0, 1, -1],
        [1, 1, -1],
        [-2, 1, -1],
        [-2, -3, -1],
        [2, 1, -1],
        [2, 1, 3],
        [-1, 1, -1],
        [0, -2, -1],
        [1, -2, -1],
        [1, 2, -1],
        [1, 2, 3],
        [-3, -2, -1],
        [-2, -2, -1],
        [-1, -2, -1],
        [0, -1, -1],
        [1, -1, -1],
        [-2, -1, -1],
        [-1, -1, -1],
    ]
    ref_bz_map = [
        0,
        1,
        2,
        4,
        5,
        6,
        7,
        8,
        9,
        11,
        12,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        34,
        35,
        36,
        40,
        41,
        43,
        44,
        46,
        47,
        48,
        49,
        50,
        54,
        56,
        57,
        59,
        60,
        61,
        65,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        77,
        78,
        79,
        83,
        84,
        85,
        86,
        87,
        88,
        89,
    ]
    ref_bzg2grg = [
        0,
        1,
        2,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        8,
        9,
        10,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        27,
        27,
        27,
        28,
        29,
        30,
        30,
        30,
        30,
        31,
        32,
        32,
        33,
        34,
        34,
        35,
        36,
        37,
        38,
        39,
        39,
        39,
        39,
        40,
        40,
        41,
        42,
        42,
        43,
        44,
        45,
        45,
        45,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        54,
        54,
        54,
        55,
        56,
        57,
        57,
        57,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
    ]
    np.testing.assert_equal(ref_bz_grid_addresses, bz_grid_addresses)
    np.testing.assert_equal(ref_bz_map, bz_map)
    np.testing.assert_equal(ref_bzg2grg, bzg2grg)
    # for adrs in bz_grid_addresses:
    #     print("[%d, %d, %d]," % tuple(adrs))
    # print(len(bz_grid_addresses))
    # print(",".join([f"%d" % x for x in bz_map]))
    # print(len(bz_map))
    # print(",".join([f"%d" % x for x in bzg2grg]))


def test_relocate_BZ_grid_address_aln_grg():
    """Test of _relocate_BZ_grid_address by wurtzite in GR-grid."""
    D_diag = [1, 5, 15]
    Q = [[-1, 0, -6], [0, -1, 0], [-1, 0, -5]]
    reciprocal_lattice = np.array(
        [
            [0.3214400514304082, 0.0, 0.0],
            [0.1855835002216734, 0.3711670004433468, 0.0],
            [0.0, 0.0, 0.20088388911209323],
        ]
    )
    bz_grid_addresses, bz_map, bzg2grg = _relocate_BZ_grid_address(
        D_diag,
        Q,
        reciprocal_lattice,  # column vectors
        PS=None,
        store_dense_gp_map=True,
    )
    ref_bz_grid_addresses = [
        [0, 0, 0],
        [0, 1, 0],
        [0, 2, 0],
        [0, -2, 0],
        [0, -1, 0],
        [0, 0, 1],
        [0, 1, 1],
        [5, 1, -14],
        [0, -3, 1],
        [5, 2, -14],
        [0, -2, 1],
        [0, -1, 1],
        [-1, 0, 2],
        [-1, 1, 2],
        [-1, 2, 2],
        [-1, -2, 2],
        [-1, 3, 2],
        [-1, -1, 2],
        [-1, 0, 3],
        [-1, 1, 3],
        [-1, 2, 3],
        [-1, -3, 3],
        [-1, -2, 3],
        [-1, -1, 3],
        [4, 0, -11],
        [4, 1, -11],
        [4, 2, -11],
        [4, 3, -11],
        [-1, -2, 4],
        [4, -1, -11],
        [-1, -1, 4],
        [-2, 0, 5],
        [-2, 1, 5],
        [-2, 2, 5],
        [-2, -2, 5],
        [-2, -1, 5],
        [-2, 0, 6],
        [-2, 1, 6],
        [3, 1, -9],
        [-2, -3, 6],
        [3, 2, -9],
        [-2, -2, 6],
        [-2, -1, 6],
        [3, 0, -8],
        [3, 1, -8],
        [3, 2, -8],
        [3, -2, -8],
        [3, 3, -8],
        [3, -1, -8],
        [-3, 0, 8],
        [-3, 1, 8],
        [-3, 2, 8],
        [-3, -3, 8],
        [-3, -2, 8],
        [-3, -1, 8],
        [2, 0, -6],
        [2, 1, -6],
        [2, 2, -6],
        [2, 3, -6],
        [-3, -2, 9],
        [2, -1, -6],
        [-3, -1, 9],
        [2, 0, -5],
        [2, 1, -5],
        [2, 2, -5],
        [2, -2, -5],
        [2, -1, -5],
        [-4, 0, 11],
        [-4, 1, 11],
        [1, 1, -4],
        [-4, -3, 11],
        [1, 2, -4],
        [-4, -2, 11],
        [-4, -1, 11],
        [1, 0, -3],
        [1, 1, -3],
        [1, 2, -3],
        [1, -2, -3],
        [1, 3, -3],
        [1, -1, -3],
        [1, 0, -2],
        [1, 1, -2],
        [1, 2, -2],
        [1, -3, -2],
        [1, -2, -2],
        [1, -1, -2],
        [0, 0, -1],
        [0, 1, -1],
        [0, 2, -1],
        [0, 3, -1],
        [-5, -2, 14],
        [0, -1, -1],
        [-5, -1, 14],
    ]
    ref_bz_map = [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        8,
        10,
        11,
        12,
        13,
        14,
        15,
        17,
        18,
        19,
        20,
        22,
        23,
        24,
        25,
        26,
        27,
        29,
        31,
        32,
        33,
        34,
        35,
        36,
        37,
        39,
        41,
        42,
        43,
        44,
        45,
        46,
        48,
        49,
        50,
        51,
        53,
        54,
        55,
        56,
        57,
        58,
        60,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        70,
        72,
        73,
        74,
        75,
        76,
        77,
        79,
        80,
        81,
        82,
        84,
        85,
        86,
        87,
        88,
        89,
        91,
        93,
    ]
    ref_bzg2grg = [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        6,
        7,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        13,
        14,
        15,
        16,
        17,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        23,
        24,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        31,
        32,
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        38,
        39,
        40,
        41,
        42,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        48,
        49,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        56,
        57,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
        63,
        64,
        65,
        66,
        67,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        73,
        74,
        74,
    ]
    np.testing.assert_equal(ref_bz_grid_addresses, bz_grid_addresses)
    np.testing.assert_equal(ref_bz_map, bz_map)
    np.testing.assert_equal(ref_bzg2grg, bzg2grg)
    # for adrs in bz_grid_addresses:
    #     print("{%d, %d, %d}," % tuple(adrs))
    # print(len(bz_grid_addresses))
    # print(",".join([f"%d" % x for x in bz_map]))
    # print(len(bz_map))
    # print(",".join([f"%d" % x for x in bzg2grg]))


def test_relocate_BZ_grid_address_aln_553():
    """Test of _relocate_BZ_grid_address by wurtzite."""
    D_diag = [5, 5, 3]
    Q = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    reciprocal_lattice = np.array(
        [
            [0.3214400514304082, 0.0, 0.0],
            [0.1855835002216734, 0.3711670004433468, 0.0],
            [0.0, 0.0, 0.20088388911209323],
        ]
    )
    bz_grid_addresses, bz_map, bzg2grg = _relocate_BZ_grid_address(
        D_diag,
        Q,
        reciprocal_lattice,  # column vectors
        PS=None,
        store_dense_gp_map=True,
    )
    ref_bz_grid_addresses = [
        [0, 0, 0],
        [1, 0, 0],
        [2, 0, 0],
        [-2, 0, 0],
        [-1, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [2, 1, 0],
        [-3, 1, 0],
        [-2, 1, 0],
        [-1, 1, 0],
        [0, 2, 0],
        [1, 2, 0],
        [1, -3, 0],
        [2, -3, 0],
        [-3, 2, 0],
        [-2, 2, 0],
        [-1, 2, 0],
        [0, -2, 0],
        [1, -2, 0],
        [2, -2, 0],
        [-2, 3, 0],
        [3, -2, 0],
        [-1, -2, 0],
        [-1, 3, 0],
        [0, -1, 0],
        [1, -1, 0],
        [2, -1, 0],
        [-2, -1, 0],
        [3, -1, 0],
        [-1, -1, 0],
        [0, 0, 1],
        [1, 0, 1],
        [2, 0, 1],
        [-2, 0, 1],
        [-1, 0, 1],
        [0, 1, 1],
        [1, 1, 1],
        [2, 1, 1],
        [-3, 1, 1],
        [-2, 1, 1],
        [-1, 1, 1],
        [0, 2, 1],
        [1, 2, 1],
        [1, -3, 1],
        [2, -3, 1],
        [-3, 2, 1],
        [-2, 2, 1],
        [-1, 2, 1],
        [0, -2, 1],
        [1, -2, 1],
        [2, -2, 1],
        [-2, 3, 1],
        [3, -2, 1],
        [-1, -2, 1],
        [-1, 3, 1],
        [0, -1, 1],
        [1, -1, 1],
        [2, -1, 1],
        [-2, -1, 1],
        [3, -1, 1],
        [-1, -1, 1],
        [0, 0, -1],
        [1, 0, -1],
        [2, 0, -1],
        [-2, 0, -1],
        [-1, 0, -1],
        [0, 1, -1],
        [1, 1, -1],
        [2, 1, -1],
        [-3, 1, -1],
        [-2, 1, -1],
        [-1, 1, -1],
        [0, 2, -1],
        [1, 2, -1],
        [1, -3, -1],
        [2, -3, -1],
        [-3, 2, -1],
        [-2, 2, -1],
        [-1, 2, -1],
        [0, -2, -1],
        [1, -2, -1],
        [2, -2, -1],
        [-2, 3, -1],
        [3, -2, -1],
        [-1, -2, -1],
        [-1, 3, -1],
        [0, -1, -1],
        [1, -1, -1],
        [2, -1, -1],
        [-2, -1, -1],
        [3, -1, -1],
        [-1, -1, -1],
    ]
    ref_bz_map = [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        9,
        10,
        11,
        12,
        14,
        16,
        17,
        18,
        19,
        20,
        21,
        23,
        25,
        26,
        27,
        28,
        30,
        31,
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        40,
        41,
        42,
        43,
        45,
        47,
        48,
        49,
        50,
        51,
        52,
        54,
        56,
        57,
        58,
        59,
        61,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        69,
        71,
        72,
        73,
        74,
        76,
        78,
        79,
        80,
        81,
        82,
        83,
        85,
        87,
        88,
        89,
        90,
        92,
        93,
    ]
    ref_bzg2grg = [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        7,
        8,
        9,
        10,
        11,
        11,
        12,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        18,
        19,
        19,
        20,
        21,
        22,
        23,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        32,
        33,
        34,
        35,
        36,
        36,
        37,
        37,
        38,
        39,
        40,
        41,
        42,
        43,
        43,
        44,
        44,
        45,
        46,
        47,
        48,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        57,
        58,
        59,
        60,
        61,
        61,
        62,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        68,
        69,
        69,
        70,
        71,
        72,
        73,
        73,
        74,
    ]
    np.testing.assert_equal(ref_bz_grid_addresses, bz_grid_addresses)
    np.testing.assert_equal(ref_bz_map, bz_map)
    np.testing.assert_equal(ref_bzg2grg, bzg2grg)
    # for adrs in bz_grid_addresses:
    #     print("{%d, %d, %d}," % tuple(adrs))
    # print(len(bz_grid_addresses))
    # print(",".join([f"%d" % x for x in bz_map]))
    # print(len(bz_map))
    # print(",".join([f"%d" % x for x in bzg2grg]))


def test_relocate_BZ_grid_address_aln_compare():
    """Test of _relocate_BZ_grid_address by zincblende.

    Compare q-points in two grids (traditional and GR grid) with the same lattice.

    """
    reciprocal_lattice = np.array(
        [
            [0.3214400514304082, 0.0, 0.0],
            [0.1855835002216734, 0.3711670004433468, 0.0],
            [0.0, 0.0, 0.20088388911209323],
        ]
    )
    D_diag = [5, 5, 3]
    Q = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    bz_grid_addresses, bz_map, bzg2grg = _relocate_BZ_grid_address(
        D_diag,
        Q,
        reciprocal_lattice,  # column vectors
        PS=None,
        store_dense_gp_map=True,
    )
    qpoints = np.dot(
        bz_grid_addresses / np.array(D_diag, dtype="double"), np.transpose(Q)
    )

    D_diag = [1, 5, 15]
    Q = [[-1, 0, -6], [0, -1, 0], [-1, 0, -5]]
    bz_grid_addresses_grg, bz_map, bzg2grg = _relocate_BZ_grid_address(
        D_diag,
        Q,
        reciprocal_lattice,  # column vectors
        PS=None,
        store_dense_gp_map=True,
    )
    qpoints_grg = np.dot(
        bz_grid_addresses_grg / np.array(D_diag, dtype="double"), np.transpose(Q)
    )

    check = [False] * 93
    for q in qpoints:
        indices = np.where(np.linalg.norm(qpoints_grg - q, axis=1) < 1e-10)
        assert len(indices[0]) == 1
        check[indices[0][0]] = True
    assert all(check)


# def test_get_aln_bzgrid(aln_cell: PhonopyAtoms):
#     """Return BZGrid of wurtzite AlN."""
#     mesh = 14
#     symmetry = Symmetry(aln_cell)
#     bzgrid = BZGrid(mesh, lattice=aln_cell.cell, symmetry_dataset=symmetry.dataset)
#     bzgrid = BZGrid(
#         mesh,
#         lattice=aln_cell.cell,
#         symmetry_dataset=symmetry.dataset,
#         use_grg=True,
#         force_SNF=True,
#     )
#     print(bzgrid.D_diag)
#     print(bzgrid.P)
#     print(bzgrid.Q)
#     shifts = np.array(
#         [
#             [0, 0, 0],
#             [0, 0, 1],
#             [0, 1, 0],
#             [0, 1, 1],
#             [1, 0, 0],
#             [1, 0, 1],
#             [1, 1, 0],
#             [1, 1, 1],
#         ]
#     )
#     print(np.dot(bzgrid.P, shifts.T).T)

#     return bzgrid


def test_GridMatrix_without_symmetry(ph_nacl: Phonopy):
    """Test of GridMatrix without symmetry."""
    mesh = [4, 4, 4]
    gm = GridMatrix(mesh, ph_nacl.primitive.cell)
    assert gm.grid_matrix is None
    np.testing.assert_array_equal(gm.D_diag, mesh)


def test_GridMatrix_with_symmetry(ph_nacl: Phonopy):
    """Test of GridMatrix with symmetry."""
    mesh = 50.0
    gm = GridMatrix(
        mesh,
        ph_nacl.primitive.cell,
        symmetry_dataset=ph_nacl.primitive_symmetry.dataset,
    )
    np.testing.assert_array_equal(
        gm.grid_matrix, np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]) * 9
    )
    np.testing.assert_array_equal(gm.D_diag, [9, 18, 18])


def test_GridMatrix_with_grid_matrix(ph_nacl: Phonopy):
    """Test of GridMatrix with grid_matrix.

    This case has to be used with symmetry_dataset or transformation_matrix.

    """
    mesh = np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]) * 9
    gm = GridMatrix(
        mesh,
        ph_nacl.primitive.cell,
        symmetry_dataset=ph_nacl.primitive_symmetry.dataset,
    )
    np.testing.assert_array_equal(gm.grid_matrix, mesh)
    np.testing.assert_array_equal(gm.D_diag, [9, 18, 18])

    tmat = ph_nacl.primitive_symmetry.dataset["transformation_matrix"]
    gm = GridMatrix(mesh, ph_nacl.primitive.cell, transformation_matrix=tmat)
    np.testing.assert_array_equal(gm.grid_matrix, mesh)
    np.testing.assert_array_equal(gm.D_diag, [9, 18, 18])


def test_GridMatrix_with_transformation_matrix(ph_nacl: Phonopy):
    """Test of GridMatrix with transformation matrix.

    Only length is used to determine grid matrix, i.e., point group information
    is not used.

    """
    mesh = 50.0
    tmat = ph_nacl.primitive_symmetry.dataset["transformation_matrix"]
    gm = GridMatrix(
        mesh,
        ph_nacl.primitive.cell,
        transformation_matrix=tmat,
    )
    np.testing.assert_array_equal(
        gm.grid_matrix, np.array([[-1, 1, 1], [1, -1, 1], [1, 1, -1]]) * 9
    )
    np.testing.assert_array_equal(gm.D_diag, [9, 18, 18])


def test_GridMatrix_with_supercell_symmetry(ph_nacl: Phonopy):
    """Test of GridMatrix with supercell symmetry.

    Generalized regular grid can not be used for non-primitive lattice.
    Therefore, fallback to length2mesh. With grg=True, warning is emitted.

    """
    mesh = 50.0
    with pytest.warns(RuntimeWarning):
        gm = GridMatrix(
            mesh,
            ph_nacl.supercell.cell,
            use_grg=True,
            symmetry_dataset=ph_nacl.symmetry.dataset,
        )
    assert gm.grid_matrix is None
    np.testing.assert_array_equal(gm.D_diag, [4, 4, 4])


@pytest.mark.filterwarnings("error")
def test_GridMatrix_with_supercell_symmetry_grg_false(ph_nacl: Phonopy):
    """Test of GridMatrix with supercell symmetry.

    With grg=False, simply length2mesh is used and warning is not emitted.

    """
    mesh = 50.0
    gm = GridMatrix(
        mesh,
        ph_nacl.supercell.cell,
        use_grg=False,
        symmetry_dataset=ph_nacl.symmetry.dataset,
    )
    assert gm.grid_matrix is None
    np.testing.assert_array_equal(gm.D_diag, [4, 4, 4])



"""Pytest conftest.py."""
from pathlib import Path

import numpy as np
import phonopy
import pytest
from phonopy import Phonopy
from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.structure.atoms import PhonopyAtoms

import phono3py
from phono3py import Phono3py

cwd = Path(__file__).parent
store_dense_gp_map = True


def pytest_addoption(parser):
    """Activate v1 emulation  with --v1 option."""
    parser.addoption(
        "--v1",
        action="store_false",
        default=True,
        help="Run with phono3py v1.x emulation.",
    )


@pytest.fixture(scope="session")
def agno2_cell() -> PhonopyAtoms:
    """Return AgNO2 cell (Imm2)."""
    cell = read_cell_yaml(cwd / "AgNO2_cell.yaml")
    return cell


@pytest.fixture(scope="session")
def aln_cell() -> PhonopyAtoms:
    """Return AlN cell (P6_3mc)."""
    a = 3.111
    c = 4.978
    lattice = [[a, 0, 0], [-a / 2, a * np.sqrt(3) / 2, 0], [0, 0, c]]
    symbols = ["Al", "Al", "N", "N"]
    positions = [
        [1.0 / 3, 2.0 / 3, 0.0009488200000000],
        [2.0 / 3, 1.0 / 3, 0.5009488200000001],
        [1.0 / 3, 2.0 / 3, 0.6190511800000000],
        [2.0 / 3, 1.0 / 3, 0.1190511800000000],
    ]
    cell = PhonopyAtoms(cell=lattice, symbols=symbols, scaled_positions=positions)
    return cell


@pytest.fixture(scope="session")
def si_pbesol(request) -> Phono3py:
    """Return Phono3py instance of Si 2x2x2.

    * with symmetry
    * full fc

    """
    yaml_filename = cwd / "phono3py_si_pbesol.yaml"
    forces_fc3_filename = cwd / "FORCES_FC3_si_pbesol"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        forces_fc3_filename=forces_fc3_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_grg(request) -> Phono3py:
    """Return Phono3py instance of Si 2x2x2.

    * with symmetry
    * full fc
    * GR-grid

    """
    yaml_filename = cwd / "phono3py_si_pbesol.yaml"
    forces_fc3_filename = cwd / "FORCES_FC3_si_pbesol"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        forces_fc3_filename=forces_fc3_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        use_grg=True,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_nosym(request) -> Phono3py:
    """Return Phono3py instance of Si 2x2x2.

    * without symmetry
    * no fc

    """
    yaml_filename = cwd / "phono3py_si_pbesol.yaml"
    forces_fc3_filename = cwd / "FORCES_FC3_si_pbesol"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        forces_fc3_filename=forces_fc3_filename,
        is_symmetry=False,
        produce_fc=False,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_nomeshsym(request) -> Phono3py:
    """Return Phono3py instance of Si 2x2x2.

    * without mesh-symmetry
    * no fc

    """
    yaml_filename = cwd / "phono3py_si_pbesol.yaml"
    forces_fc3_filename = cwd / "FORCES_FC3_si_pbesol"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        forces_fc3_filename=forces_fc3_filename,
        is_mesh_symmetry=False,
        produce_fc=False,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_compact_fc(request) -> Phono3py:
    """Return Phono3py instance of Si 2x2x2.

    * with symmetry
    * compact fc

    """
    yaml_filename = cwd / "phono3py_si_pbesol.yaml"
    forces_fc3_filename = cwd / "FORCES_FC3_si_pbesol"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        forces_fc3_filename=forces_fc3_filename,
        is_compact_fc=True,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc

    """
    yaml_filename = cwd / "phono3py_params_Si111.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_alm(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm if available on test side

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si111.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="alm",
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_iterha_111() -> Phonopy:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * no fc

    """
    yaml_filename = cwd / "phonopy_params-Si111-iterha.yaml.gz"
    return phonopy.load(yaml_filename, log_level=1, produce_fc=False)


@pytest.fixture(scope="session")
def si_pbesol_111_222_fd(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm if available on test side

    """
    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_222_alm(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm if available on test side

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="alm",
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_222_alm_fd(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm for fc2 if available on test side

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="alm|",
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_222_fd_alm(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm for fc3 if available on test side

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="|alm",
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_222_alm_cutoff(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm if available on test side
    * cutoff=3

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="alm",
        fc_calculator_options="cutoff = 3",
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_222_alm_cutoff_fc2(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm if available on test side
    * cutoff=3 only for fc2

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="alm",
        fc_calculator_options="cutoff = 3|",
        log_level=1,
    )


@pytest.fixture(scope="session")
def si_pbesol_111_222_alm_cutoff_fc3(request) -> Phono3py:
    """Return Phono3py instance of Si 1x1x1.

    * with symmetry
    * full fc
    * use alm if available on test side
    * cutoff=3 only for fc3

    """
    pytest.importorskip("alm")

    yaml_filename = cwd / "phono3py_params_Si-111-222.yaml"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        fc_calculator="alm",
        fc_calculator_options="|cutoff = 3",
        log_level=1,
    )


@pytest.fixture(scope="session")
def nacl_pbe(request) -> Phono3py:
    """Return Phono3py instance of NaCl 2x2x2.

    * with symmetry
    * full fc

    """
    yaml_filename = cwd / "phono3py_params_NaCl222.yaml.xz"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def nacl_pbe_compact_fc(request) -> Phono3py:
    """Return Phono3py instance of NaCl 2x2x2.

    * with symmetry
    * compact fc

    """
    yaml_filename = cwd / "phono3py_params_NaCl222.yaml.xz"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        is_compact_fc=True,
        log_level=1,
    )


@pytest.fixture(scope="session")
def nacl_pbe_cutoff_fc3(request) -> Phono3py:
    """Return Phono3py instance of NaCl 2x2x2 with cutoff-pair-distance.

    * cutoff pair with 5

    """
    yaml_filename = cwd / "phono3py_params_NaCl222.yaml.xz"
    enable_v2 = request.config.getoption("--v1")
    ph3 = phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        produce_fc=False,
        log_level=1,
    )
    forces = ph3.forces
    ph3.generate_displacements(cutoff_pair_distance=5)
    dataset = ph3.dataset
    dataset["first_atoms"][0]["forces"] = forces[0]
    dataset["first_atoms"][1]["forces"] = forces[1]
    count = 2
    for first_atoms in dataset["first_atoms"]:
        for second_atoms in first_atoms["second_atoms"]:
            assert second_atoms["id"] == count + 1
            second_atoms["forces"] = forces[count]
            count += 1
    ph3.dataset = dataset
    ph3.produce_fc3()
    # ph3.produce_fc3(symmetrize_fc3r=True)
    return ph3


@pytest.fixture(scope="session")
def nacl_pbe_cutoff_fc3_all_forces(request) -> Phono3py:
    """Return Phono3py instance of NaCl 2x2x2 with cutoff-pair-distance.

    * cutoff pair with 5
    * All forces are set.

    """
    yaml_filename = cwd / "phono3py_params_NaCl222.yaml.xz"
    enable_v2 = request.config.getoption("--v1")
    ph3 = phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        produce_fc=False,
        log_level=1,
    )
    forces = ph3.forces
    ph3.generate_displacements(cutoff_pair_distance=5)
    ph3.forces = forces
    ph3.produce_fc3()
    return ph3


@pytest.fixture(scope="session")
def nacl_pbe_cutoff_fc3_compact_fc(request) -> Phono3py:
    """Return Phono3py instance of NaCl 2x2x2 with cutoff-pair-distance.

    * cutoff pair with 5
    * All forces are set.
    * Compact FC

    """
    yaml_filename = cwd / "phono3py_params_NaCl222.yaml.xz"
    enable_v2 = request.config.getoption("--v1")
    ph3 = phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        produce_fc=False,
        log_level=1,
    )
    forces = ph3.forces
    ph3.generate_displacements(cutoff_pair_distance=5)
    ph3.forces = forces
    ph3.produce_fc3(is_compact_fc=True)
    return ph3


@pytest.fixture(scope="session")
def aln_lda(request) -> Phono3py:
    """Return Phono3py instance of AlN 3x3x2.

    * with symmetry
    * full fc.

    """
    yaml_filename = cwd / "phono3py_params_AlN332.yaml.xz"
    enable_v2 = request.config.getoption("--v1")
    return phono3py.load(
        yaml_filename,
        store_dense_gp_map=enable_v2,
        store_dense_svecs=enable_v2,
        log_level=1,
    )


@pytest.fixture(scope="session")
def ph_nacl() -> Phonopy:
    """Return Phonopy class instance of NaCl 2x2x2."""
    yaml_filename = cwd / "phonopy_disp_NaCl.yaml"
    force_sets_filename = cwd / "FORCE_SETS_NaCl"
    born_filename = cwd / "BORN_NaCl"
    return phonopy.load(
        yaml_filename,
        force_sets_filename=force_sets_filename,
        born_filename=born_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )


@pytest.fixture(scope="session")
def ph_si() -> Phonopy:
    """Return Phonopy class instance of Si-prim 2x2x2."""
    yaml_filename = cwd / "phonopy_params_Si.yaml"
    return phonopy.load(
        yaml_filename,
        is_compact_fc=False,
        log_level=1,
        produce_fc=True,
    )



"""Phono3py setup.py.

Cmake handles automatic finding of library.
Custom settings should be written in site.cfg.

To fully customize using site.cfg,
set PHONO3PY_USE_CMAKE=false to avoid running cmake.

"""
import os
import pathlib
import shutil
import subprocess

import numpy
import setuptools

if (
    "PHONO3PY_USE_CMAKE" in os.environ
    and os.environ["PHONO3PY_USE_CMAKE"].lower() == "false"
):
    use_cmake = False
else:
    use_cmake = True


def _run_cmake(build_dir):
    build_dir.mkdir()
    args = [
        "cmake",
        "-S",
        ".",
        "-B",
        "_build",
        "-DPHONONCALC=on",
        "-DPHONO3PY=on",
        "-DCMAKE_INSTALL_PREFIX=.",
    ]
    # if "CONDA_PREFIX" in os.environ:
    #     args.append("-DUSE_CONDA_PATH=on")
    # if "CC" in os.environ:
    #     args.append(f'-DCMAKE_C_COMPILER={os.environ["CC"]}')

    cmake_output = subprocess.check_output(args)
    print(cmake_output.decode("utf-8"))
    subprocess.check_call(["cmake", "--build", "_build", "-v"])
    return cmake_output


def _clean_cmake(build_dir):
    if build_dir.exists():
        shutil.rmtree(build_dir)


def _get_params_from_site_cfg():
    """Read extra_compile_args and extra_link_args.

    Examples
    --------
    # For macOS
    extra_compile_args = -fopenmp=libomp
    extra_link_args = -lomp -lopenblas

    # For linux
    extra_compile_args = -fopenmp
    extra_link_args = -lgomp  -lopenblas -lpthread

    """
    params = {
        "define_macros": [],
        "extra_link_args": [],
        "extra_compile_args": [],
        "extra_objects": [],
        "include_dirs": [],
    }
    use_mkl_lapacke = False
    use_threaded_blas = False

    site_cfg_file = pathlib.Path.cwd() / "site.cfg"
    if not site_cfg_file.exists():
        return params

    with open(site_cfg_file) as f:
        lines = [line.strip().split("=", maxsplit=1) for line in f]

        for line in lines:
            if len(line) < 2:
                continue
            key = line[0].strip()
            val = line[1].strip()
            if key not in params:
                continue
            if key == "define_macros":
                pair = val.split(maxsplit=1)
                if pair[1].lower() == "none":
                    pair[1] = None
                params[key].append(tuple(pair))
            else:
                if "mkl" in val:
                    use_mkl_lapacke = True
                if "openblas" in val:
                    use_threaded_blas = True
                params[key] += val.split()

    if use_mkl_lapacke:
        params["define_macros"].append(("MKL_LAPACKE", None))
    if use_threaded_blas:
        params["define_macros"].append(("MULTITHREADED_BLAS", None))
    if "THM_EPSILON" not in [macro[0] for macro in params["define_macros"]]:
        params["define_macros"].append(("THM_EPSILON", "1e-10"))

    print("=============================================")
    print("Parameters found in site.cfg")
    for key, val in params.items():
        print(f"{key}: {val}")
    print("=============================================")
    return params


def _get_extensions(build_dir):
    """Return python extension setting.

    User customization by site.cfg file
    -----------------------------------
    See _get_params_from_site_cfg().

    Automatic search using cmake
    ----------------------------
    Invoked by environment variable unless PHONO3PY_USE_CMAKE=false.

    """
    params = _get_params_from_site_cfg()
    extra_objects_ph3py = []
    extra_objects_phcalc = []

    if not use_cmake or not shutil.which("cmake"):
        print("** Setup without using cmake **")
        sources_ph3py = [
            "c/_phono3py.c",
            "c/bzgrid.c",
            "c/collision_matrix.c",
            "c/fc3.c",
            "c/grgrid.c",
            "c/imag_self_energy_with_g.c",
            "c/interaction.c",
            "c/isotope.c",
            "c/lagrid.c",
            "c/lapack_wrapper.c",
            "c/phono3py.c",
            "c/phonoc_utils.c",
            "c/pp_collision.c",
            "c/real_self_energy.c",
            "c/real_to_reciprocal.c",
            "c/reciprocal_to_normal.c",
            "c/snf3x3.c",
            "c/tetrahedron_method.c",
            "c/triplet.c",
            "c/triplet_grid.c",
            "c/triplet_iw.c",
        ]
        sources_phcalc = [
            "c/_phononcalc.c",
            "c/dynmat.c",
            "c/lapack_wrapper.c",
            "c/phonon.c",
            "c/phononcalc.c",
        ]
    else:
        print("** Setup using cmake **")
        use_mkl_lapacke = False
        found_extra_link_args = []
        found_extra_compile_args = []
        sources_ph3py = ["c/_phono3py.c"]
        sources_phcalc = ["c/_phononcalc.c"]
        cmake_output = _run_cmake(build_dir)
        found_flags = {}
        found_libs = {}
        for line in cmake_output.decode("utf-8").split("\n"):
            for key in ["BLAS", "LAPACK", "OpenMP"]:
                if f"{key} libs" in line and len(line.split()) > 3:
                    found_libs[key] = line.split()[3].split(";")
                if f"{key} flags" in line and len(line.split()) > 3:
                    found_flags[key] = line.split()[3].split(";")
        for key, value in found_libs.items():
            found_extra_link_args += value
            for element in value:
                if "libmkl" in element:
                    use_mkl_lapacke = True
        for key, value in found_flags.items():
            found_extra_compile_args += value
        if use_mkl_lapacke:
            params["define_macros"].append(("MKL_LAPACKE", None))

        libph3py = list((pathlib.Path.cwd() / "_build").glob("*ph3py.*"))
        if libph3py:
            print("=============================================")
            print(f"Phono3py library: {libph3py[0]}")
            print("=============================================")
            extra_objects_ph3py += [str(libph3py[0])]

        libphcalc = list((pathlib.Path.cwd() / "_build").glob("*phcalc.*"))
        if libphcalc:
            print("=============================================")
            print(f"Phonon library: {libphcalc[0]}")
            print("=============================================")
            extra_objects_phcalc += [str(libphcalc[0])]

        params["extra_link_args"] += found_extra_link_args
        params["extra_compile_args"] += found_extra_compile_args

        print("=============================================")
        print("Parameters found by cmake")
        print("extra_compile_args: ", found_extra_compile_args)
        print("extra_link_args: ", found_extra_link_args)
        print("define_macros: ", params["define_macros"])
        print("=============================================")
        print()

    extensions = []
    params["define_macros"].append(("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION"))
    params["include_dirs"] += ["c", numpy.get_include()]

    extensions.append(
        setuptools.Extension(
            "phono3py._phono3py",
            sources=sources_ph3py,
            extra_link_args=params["extra_link_args"],
            include_dirs=params["include_dirs"],
            extra_compile_args=params["extra_compile_args"],
            extra_objects=params["extra_objects"] + extra_objects_ph3py,
            define_macros=params["define_macros"],
        )
    )

    extensions.append(
        setuptools.Extension(
            "phono3py._phononcalc",
            sources=sources_phcalc,
            extra_link_args=params["extra_link_args"],
            include_dirs=params["include_dirs"],
            extra_compile_args=params["extra_compile_args"],
            extra_objects=params["extra_objects"] + extra_objects_phcalc,
            define_macros=params["define_macros"],
        )
    )
    return extensions


def _get_version() -> str:
    git_num = None
    version_nums = [None, None, None]
    with open("phono3py/version.py") as w:
        for line in w:
            if "__version__" in line:
                for i, num in enumerate(line.split()[2].strip('"').split(".")):
                    version_nums[i] = num
                break

    # To deploy to pypi by travis-CI
    if os.path.isfile("__nanoversion__.txt"):
        nanoversion = 0
        with open("__nanoversion__.txt") as nv:
            try:
                for line in nv:
                    nanoversion = int(line.strip())
                    break
            except ValueError:
                nanoversion = 0
        if nanoversion != 0:
            version_nums.append(nanoversion)
    elif git_num:
        version_nums.append(git_num)

    if None in version_nums:
        print("Failed to get version number in setup.py.")
        raise

    version = ".".join(["%s" % n for n in version_nums[:3]])
    if len(version_nums) > 3:
        version += "-%s" % version_nums[3]
    return version


def main(build_dir):
    """Run setuptools."""
    version = _get_version()

    packages_phono3py = [
        "phono3py",
        "phono3py.conductivity",
        "phono3py.cui",
        "phono3py.interface",
        "phono3py.other",
        "phono3py.phonon",
        "phono3py.phonon3",
        "phono3py.sscha",
    ]
    scripts_phono3py = [
        "scripts/phono3py",
        "scripts/phono3py-load",
        "scripts/phono3py-kaccum",
        "scripts/phono3py-kdeplot",
        "scripts/phono3py-coleigplot",
    ]

    setuptools.setup(
        name="phono3py",
        version=version,
        description="This is the phono3py module.",
        author="Atsushi Togo",
        author_email="atz.togo@gmail.com",
        url="http://phonopy.github.io/phono3py/",
        packages=packages_phono3py,
        python_requires=">=3.8",
        install_requires=[
            "numpy>=1.17.0",
            "scipy",
            "PyYAML>=5.3",
            "matplotlib>=2.2.2",
            "h5py>=3.0",
            "spglib>=2.0",
            "phonopy>=2.20,<2.21",
        ],
        provides=["phono3py"],
        scripts=scripts_phono3py,
        ext_modules=_get_extensions(build_dir),
    )
    _clean_cmake(build_dir)


if __name__ == "__main__":
    build_dir = pathlib.Path.cwd() / "_build"
    main(build_dir)



"""API for joint-density-of-states calculation."""
# Copyright (C) 2019 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np
from phonopy.structure.cells import Primitive, Supercell
from phonopy.structure.symmetry import Symmetry
from phonopy.units import VaspToTHz

from phono3py.file_IO import write_joint_dos
from phono3py.phonon3.imag_self_energy import (
    get_freq_points_batches,
    get_frequency_points,
)
from phono3py.phonon3.joint_dos import JointDos
from phono3py.phonon.grid import BZGrid


class Phono3pyJointDos:
    """Class to calculate joint-density-of-states."""

    def __init__(
        self,
        supercell: Supercell,
        primitive: Primitive,
        fc2,
        mesh=None,
        nac_params=None,
        nac_q_direction=None,
        sigmas=None,
        cutoff_frequency=1e-4,
        frequency_step=None,
        num_frequency_points=None,
        num_points_in_batch=None,
        temperatures=None,
        frequency_factor_to_THz=VaspToTHz,
        frequency_scale_factor=None,
        use_grg=False,
        SNF_coordinates="reciprocal",
        is_mesh_symmetry=True,
        is_symmetry=True,
        store_dense_gp_map=True,
        symprec=1e-5,
        output_filename=None,
        log_level=0,
    ):
        """Init method."""
        self._primitive = primitive
        self._supercell = supercell
        self._fc2 = fc2
        self._temperatures = temperatures
        self._nac_params = nac_params
        self._nac_q_direction = nac_q_direction
        if sigmas is None:
            self._sigmas = [None]
        else:
            self._sigmas = sigmas
        self._cutoff_frequency = cutoff_frequency
        self._frequency_factor_to_THz = frequency_factor_to_THz
        self._frequency_scale_factor = frequency_scale_factor
        self._is_mesh_symmetry = is_mesh_symmetry
        self._is_symmetry = is_symmetry

        if not store_dense_gp_map:
            warnings.warn(
                "Phono3pyJointDos init parameter of store_dense_gp_map is deprecated. "
                "This will be set always True.",
                DeprecationWarning,
            )
        self._store_dense_gp_map = store_dense_gp_map

        self._use_grg = use_grg
        self._SNF_coordinates = SNF_coordinates
        self._symprec = symprec
        self._filename = output_filename
        self._log_level = log_level

        self._bz_grid = None
        self._joint_dos = None
        self._num_frequency_points_in_batch = num_points_in_batch
        self._frequency_step = frequency_step
        self._num_frequency_points = num_frequency_points

        self._primitive_symmetry = Symmetry(
            self._primitive, self._symprec, self._is_symmetry
        )

        if mesh is not None:
            self.mesh_numbers = mesh
            self.initialize(mesh)

    @property
    def grid(self):
        """Return BZGrid class instance."""
        return self._bz_grid

    @property
    def nac_params(self):
        """Setter and getter of parameters for non-analytical term correction."""
        return self._nac_params

    @property
    def num_frequency_points_in_batch(self):
        """Getter and setter of num_frequency_points_in_batch.

        Number of sampling frequency points per batch.
        Larger value gives better concurrency in tetrahedron method,
        but requires more memory.

        """
        return self._num_frequency_points_in_batch

    @num_frequency_points_in_batch.setter
    def num_frequency_points_in_batch(self, nelems_in_batch):
        self._num_frequency_points_in_batch = nelems_in_batch

    @property
    def mesh_numbers(self):
        """Setter and getter of sampling mesh numbers in reciprocal space."""
        if self._bz_grid is None:
            return None
        else:
            return self._bz_grid.D_diag

    @mesh_numbers.setter
    def mesh_numbers(self, mesh_numbers):
        self._bz_grid = BZGrid(
            mesh_numbers,
            lattice=self._primitive.cell,
            symmetry_dataset=self._primitive_symmetry.dataset,
            is_time_reversal=self._is_symmetry,
            use_grg=self._use_grg,
            force_SNF=False,
            SNF_coordinates=self._SNF_coordinates,
            store_dense_gp_map=self._store_dense_gp_map,
        )

    def initialize(self, mesh_numbers):
        """Initialize JointDos."""
        self._jdos = JointDos(
            self._primitive,
            self._supercell,
            self._bz_grid,
            self._fc2,
            nac_params=self._nac_params,
            cutoff_frequency=self._cutoff_frequency,
            frequency_factor_to_THz=self._frequency_factor_to_THz,
            frequency_scale_factor=self._frequency_scale_factor,
            is_mesh_symmetry=self._is_mesh_symmetry,
            store_dense_gp_map=self._store_dense_gp_map,
            symprec=self._symprec,
            filename=self._filename,
            log_level=self._log_level,
        )
        if self._log_level:
            print("Generating grid system ... ", end="", flush=True)

        self.mesh_numbers = mesh_numbers

        if self._log_level:
            if self._bz_grid.grid_matrix is None:
                print("[ %d %d %d ]" % tuple(self._bz_grid.D_diag))
            else:
                print("")
                print(
                    "Generalized regular grid: [ %d %d %d ]"
                    % tuple(self._bz_grid.D_diag)
                )
                print("Grid generation matrix:")
                print("  [ %d %d %d ]" % tuple(self._bz_grid.grid_matrix[0]))
                print("  [ %d %d %d ]" % tuple(self._bz_grid.grid_matrix[1]))
                print("  [ %d %d %d ]" % tuple(self._bz_grid.grid_matrix[2]))

    def run(self, grid_points, write_jdos=False):
        """Calculate joint-density-of-states."""
        if self._log_level:
            print(
                "--------------------------------- Joint DOS "
                "---------------------------------"
            )
            print("Running harmonic phonon calculations...", flush=True)

        self._jdos.run_phonon_solver()
        frequencies, _, _ = self._jdos.get_phonons()
        self._jdos.run_phonon_solver_at_gamma()
        max_phonon_freq = np.max(frequencies)
        self._jdos.run_phonon_solver_at_gamma(is_nac=True)

        self._frequency_points = get_frequency_points(
            max_phonon_freq=max_phonon_freq,
            sigmas=self._sigmas,
            frequency_points=None,
            frequency_step=self._frequency_step,
            num_frequency_points=self._num_frequency_points,
        )
        batches = get_freq_points_batches(
            len(self._frequency_points), nelems=self._num_frequency_points_in_batch
        )
        if self._temperatures is None:
            temperatures = [None]
        else:
            temperatures = self._temperatures
        self._joint_dos = np.zeros(
            (
                len(self._sigmas),
                len(temperatures),
                len(self._frequency_points),
                2,
            ),
            dtype="double",
            order="C",
        )

        for i, gp in enumerate(grid_points):
            if (self._bz_grid.addresses[gp] == 0).all():
                self._jdos.nac_q_direction = self._nac_q_direction
            else:
                self._jdos.nac_q_direction = None
            self._jdos.set_grid_point(gp)

            if self._log_level:
                weights = self._jdos.get_triplets_at_q()[1]
                print(
                    "======================= "
                    "Grid point %d (%d/%d) "
                    "=======================" % (gp, i + 1, len(grid_points))
                )
                adrs = self._jdos.bz_grid.addresses[gp]
                q = np.dot(adrs, self._bz_grid.QDinv.T)
                print("q-point: (%5.2f %5.2f %5.2f)" % tuple(q))
                print("Number of triplets: %d" % len(weights))
                print("Frequency")
                for f in self._jdos.get_phonons()[0][gp]:
                    print("%8.3f" % f)

            if not self._sigmas:
                raise RuntimeError("sigma or tetrahedron method has to be set.")

            for i_s, sigma in enumerate(self._sigmas):
                self._jdos.sigma = sigma
                if self._log_level:
                    if sigma is None:
                        print("Tetrahedron method is used.")
                    else:
                        print("Smearing method with sigma=%s is used." % sigma)
                    print(
                        f"Calculations at {len(self._frequency_points)} "
                        f"frequency points are devided into {len(batches)} batches."
                    )
                for i_t, temperature in enumerate(temperatures):
                    self._jdos.temperature = temperature

                    for ib, freq_indices in enumerate(batches):
                        if self._log_level:
                            print(
                                f"{ib + 1}/{len(batches)}: {freq_indices + 1}",
                                flush=True,
                            )
                        self._jdos.frequency_points = self._frequency_points[
                            freq_indices
                        ]
                        self._jdos.run()
                        self._joint_dos[i_s, i_t, freq_indices] = self._jdos.joint_dos

                    if write_jdos:
                        filename = self._write(gp, i_sigma=i_s)
                        if self._log_level:
                            print('JDOS is written into "%s".' % filename)

    @property
    def dynamical_matrix(self):
        """Return DynamicalMatrix class instance."""
        return self._jdos.dynamical_matrix

    @property
    def frequency_points(self):
        """Return frequency points."""
        return self._frequency_points

    @property
    def joint_dos(self):
        """Return calculated joint-density-of-states."""
        return self._joint_dos

    def _write(self, gp, i_sigma=0):
        return write_joint_dos(
            gp,
            self._bz_grid.D_diag,
            self._frequency_points,
            self._joint_dos[i_sigma],
            sigma=self._sigmas[i_sigma],
            temperatures=self._temperatures,
            filename=self._filename,
            is_mesh_symmetry=self._is_mesh_symmetry,
        )



"""Tests for direct solution of LBTE."""
import numpy as np

from phono3py.api_phono3py import Phono3py

si_pbesol_kappa_P_LBTE = [111.123, 111.123, 111.123, 0, 0, 0]  # old value 111.117
si_pbesol_kappa_C = [0.167, 0.167, 0.167, 0.000, 0.000, 0.000]

si_pbesol_kappa_P_LBTE_redcol = [62.783, 62.783, 62.783, 0, 0, 0]  # old value 63.019
si_pbesol_kappa_C_redcol = (
    -1
)  # coherences conductivity is not implemented for is_reducible_collision_matrix=True,


def test_kappa_LBTE(si_pbesol: Phono3py):
    """Test for symmetry reduced collision matrix."""
    si_pbesol.mesh_numbers = [9, 9, 9]
    si_pbesol.init_phph_interaction()
    si_pbesol.run_thermal_conductivity(
        is_LBTE=True,
        temperatures=[
            300,
        ],
        conductivity_type="wigner",
    )
    # kappa = si_pbesol.thermal_conductivity.kappa.ravel()
    kappa_P = si_pbesol.thermal_conductivity.kappa_P_exact.ravel()
    kappa_C = si_pbesol.thermal_conductivity.kappa_C.ravel()
    np.testing.assert_allclose(si_pbesol_kappa_P_LBTE, kappa_P, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C, kappa_C, atol=0.02)


'''
#coherences conductivity is not implemented for is_reducible_collision_matrix=True,
def test_kappa_LBTE_full_colmat(si_pbesol: Phono3py):
    """Test for full collision matrix."""
    si_pbesol.mesh_numbers = [5, 5, 5]
    si_pbesol.init_phph_interaction()
    si_pbesol.run_thermal_conductivity(
        is_LBTE=True,
        temperatures=[
            300,
        ],
        is_reducible_collision_matrix=True,
    )
    kappa_P = si_pbesol.thermal_conductivity.kappa_P_exact.ravel()
    kappa_C = si_pbesol.thermal_conductivity.kappa_C.ravel()
    np.testing.assert_allclose(si_pbesol_kappa_P_LBTE_redcol, kappa_P, atol=0.5)
    np.testing.assert_allclose(si_pbesol_kappa_C_redcol, kappa_C, atol=0.02)
'''
